from io import StringIO
from typing import Any

import pandas as pd
from bs4 import NavigableString
from markdownify import MarkdownConverter  # type: ignore : no stubs

from .logger import LOGGER


class CustomMarkdownConverter(MarkdownConverter):
    """A custom MarkdownConverter that handles rowspan (merged cells).

    This class is a subclass of the MarkdownConverter class from the markdownify library.
    It overrides the convert_table, convert_th, convert_tr, convert_td, convert_thead, and convert_tbody methods
    to provide a No-Op for the <th>, <tr>, <td>, <thead>, and <tbody> tags, respectively.
    For <table> tags, it converts the table to a DataFrame and then converts the DataFrame to Markdown.

    It requires pandas and tabula, which is probably why markdownify doesn't implement this natively.
    As we do have both in chunknorris, we can afford using them here to improve markdown conversion of table.
    """

    def convert_table(self, el: str, text: str, parent_tags: Any) -> str:
        try:
            df = pd.read_html(StringIO(str(el)), displayed_only=False)[0]  # type: ignore : missing typing in pandas
            df = df.fillna("")  # type: ignore : missing typing in pandas
            # if no table header -> set first row as header
            if all(df.columns == list(range(len(df.columns)))):
                df.columns = df.iloc[0]
                df = df.iloc[1:]
            return "\n\n" + df.to_markdown(index=False) + "\n\n"
        except Exception as e:
            LOGGER.warning(f"Error converting table: {e}")
            return "\n\n"

    def convert_th(self, el: NavigableString, text: str, parent_tags: Any) -> str:
        """This method is empty because we want a No-Op for the <th> tag."""
        # return the html as is
        return str(el)

    def convert_tr(self, el: NavigableString, text: str, parent_tags: Any) -> str:
        """This method is empty because we want a No-Op for the <tr> tag."""
        return str(el)

    def convert_td(self, el: NavigableString, text: str, parent_tags: Any) -> str:
        """This method is empty because we want a No-Op for the <td> tag."""
        return str(el)



import re
from collections import Counter
from itertools import groupby

from thefuzz import fuzz  # type: ignore : no stubs

from .components import TocTitle
from .utils import PdfParserState


class PdfTocExtraction(PdfParserState):
    """Class intended to be used to extract the table of content
    of a document.
    Intended to be a component inherited by pdfParser => PdfParser(PdfTocExtraction)
    """

    @property
    def header_patterns(self) -> list[tuple[int, re.Pattern[str]]]:
        """A list of tuple (level, pattern) used to match the potential headers"""
        return [
            (3, re.compile(r"^(\d+)[\s\.\)]+(\d+)[\s\.\)]+(\d+)")),
            (2, re.compile(r"^(\d+)[\s\.\)]+(\d+)")),
            (1, re.compile(r"^(\d+)[\s\.\)]*")),
        ]

    def get_toc(self) -> list[TocTitle]:
        """Gets the table of content of a document.
        The process:
        - tries to find the toc in metadatas
        - if not found, tries to find a table of content in the document
        --- if found, tries to infer the level of header...
             --- ...with the x offset the headers in the toc, assuming lower level header are more to the right
             --- ...or trying to parse a schema, such as 1. 1.1, 1.1.a)
        - if not found, tries to infer header level with font size, assuming lower level have lower font size

        Returns:
            list[TocTitle]: the list of titles found
        """
        toc_from_metadata = self.get_toc_from_metadata()
        # Run anyway as table of content might also be in document
        # and we want to flag the lines that belong to it
        toc_from_document = self.get_toc_from_document()

        toc = None
        if toc_from_metadata:
            self._set_block_issectiontitle_with_toc(toc_from_metadata)
            # sometime, the toc in metadata doesn't represent the toc in document.
            # So, if less than half of toc are found in doc, find toc in doc.
            if self.headers_have_been_found(toc_from_metadata):
                toc = toc_from_metadata
        if toc is None and toc_from_document:
            self._set_block_issectiontitle_with_toc(toc_from_document)
            if self.headers_have_been_found(toc_from_document):
                toc = toc_from_document
        if toc is None:
            toc = self._set_block_issectiontitle_with_fontsize()

        return toc or []

    def get_toc_from_metadata(self) -> list[TocTitle]:
        """Uses pymupdf.get_toc() to try to get the table of
        content of the document from the metadatas.
        If it exists, builds the table of content from it

        Returns:
            list[TocTitle] | None: The toc titles found in metadatas
        """
        toc = self.document.get_toc() if self.document is not None else []  # type: ignore : missing typing in pymupdf | Document.get_toc() -> list[tuple[int, str, int, dict[str, Any]]]
        if not toc:
            return []

        return [
            TocTitle(text=item[1], level=item[0], page=item[2], source="metadata")
            for item in toc
        ]

    def get_toc_from_document(self) -> list[TocTitle]:
        """Tries to find a table of content in the document
        and parses it's titles. May return an ampty list
        if no table of content found

        Returns:
            list[TocTitle]: the titles found in the toc
        """
        toc = self._find_toc_titles()
        if toc:
            toc = self._infer_level_with_schema(toc)
            if not all(t.level for t in toc):
                toc = PdfTocExtraction._infer_level_with_offset(toc)

        return toc

    def _find_toc_titles(self) -> list[TocTitle] | None:
        """Tries to find the table of content based on regex matches.
        Handles the case where the title of the toc is on multiple lines
        Returns:
            list[TocTitle]: The potential titles of the table of content
                and their caracteristics
        """
        toc_pattern = re.compile(
            r"(.+?)(?:\s+)?[\.\_\-\s….]{5,}(?:\s+)?(?:[pP]\.\s*)?(\d+)"
        )

        toc_titles: list[TocTitle] = []
        until_last_match_counter = 0
        # browse through lines up to page 15 to get those which might be TOC
        for i, line in enumerate(filter(lambda x: x.page < 15, self.lines)):
            until_last_match_counter += 1
            if len(toc_titles) > 3 and until_last_match_counter > 10:
                break  # likely we have found a TOC and are now browsing through document
            match = re.match(toc_pattern, line.text)
            # regex may match ZIP codes or phone numbers => check page is less than 3 numbers
            if match and len(match[2]) < 4:
                until_last_match_counter = 0
                line.is_toc_element = True
                # multiline toc title
                if i > 2 and (
                    not self.lines[i - 1].is_toc_element
                    and self.lines[i - 2].is_toc_element
                ):
                    toc_text = self.lines[i - 1].text + match[1]
                    x_offset = self.lines[i - 1].origin.x  # type: ignore : missing typing in pymupdf | Point.x : float
                    self.lines[i - 1].is_toc_element = True
                # single line toc title
                else:
                    toc_text = match[1]
                    x_offset = line.origin.x  # type: ignore : missing typing in pymupdf | Point.x : float
                toc_titles.append(
                    TocTitle(
                        text=toc_text,
                        page=int(match[2]),
                        x_offset=int(x_offset),
                        source="regex",
                        source_page=line.page,
                    )
                )

        return toc_titles

    @staticmethod
    def _infer_level_with_offset(toc_titles: list[TocTitle]) -> list[TocTitle]:
        """Given on a list of toc_tiles found in the document,
        infer their level (h1, h2, ...) based on the x_offset
        of their bbox. We assume that a bigger x_offset means lower level
        (= text is more on the right of the document). Like so:
        A. H1 title
          a. H2 title
            1. H3 title

        Args:
            toc_titles (list[TocTitle]): list of toc titles returned by
                _find_toc_titles()

        Returns:
            list[TocTitle]: the toc titles, with the "level" value set
        """
        new_toc_titles: list[TocTitle] = []
        toc_elem_per_page = groupby(toc_titles, key=lambda x: x.source_page)
        for _, toc_elem_on_page in toc_elem_per_page:
            toc_elem_on_page = list(toc_elem_on_page)
            offsets: list[float] = sorted(
                list(set(t.x_offset for t in toc_elem_on_page))
            )
            for title in toc_elem_on_page:
                title.level = offsets.index(title.x_offset) + 1
                new_toc_titles.append(title)

        return new_toc_titles

    def _infer_level_with_schema(self, toc_titles: list[TocTitle]) -> list[TocTitle]:
        """Tries to infer the titles levels based on the
        schema of the title sections.

        Args:
            toc_titles (list[TocTitle]): The list of toc title

        Returns:
            list[TocTitle]: the list of toc title with the level set (if schema found)
        """
        for t in toc_titles:
            t.level = self._get_header_level_using_schema(t.text)

        return toc_titles

    def _get_header_level_using_schema(self, header_text: str) -> int | None:
        """Detects the header level according to specific schema such as :
        1. H1 title
        1.1 H2 title
        1.1.1 H3 title

        Args:
            header_text (str): the text to get the header level from

        Returns:
            int | None: the header level. Returns None if the regex didn't match any level
        """
        for level, pattern in self.header_patterns:
            if re.match(pattern, header_text):
                return level

    def _set_block_issectiontitle_with_toc(self, toc: list[TocTitle]) -> None:
        """After finding the toc, this modifies
        blocks among self.blocks attributes of the PdfParser
        to set which ones are the section's headers.
        """
        for title in toc:
            pages_to_look_on = range(title.page - 1, title.page + 2)
            blocks_to_consider = (
                block for block in self.blocks if block.page in pages_to_look_on
            )
            best_ratio, best_block = 0, None
            for block in blocks_to_consider:
                # if the first line of the block corresponds to the title text -> toc title
                if block.lines[0].text.lower() == title.text.lower():
                    block.lines[0].is_toc_element = True
                    best_ratio = 100
                    best_block = block
                    break
                # else check if title text is similar to block text
                ratio = fuzz.ratio(  # type: ignore : missing typing in fuzz | fuzz.ratio(s1 : str, s2: str) -> int
                    title.text.lower(), block.text.lower()
                )
                if ratio > best_ratio:
                    best_ratio = ratio
                    best_block = block
            if best_block and best_ratio >= 75:
                title.found = True
                best_block.section_title = title

    def _set_block_issectiontitle_with_fontsize(self) -> list[TocTitle]:
        """Uses the fontize attribute of lines to assign
        them a header level to blocks.
        We assume that bigger fontsize means higher level header.
        All block having a fontsize greater than the fontsize of body
        will be assigned a level based on their fontsize.
        """
        toc: list[TocTitle] = []
        # Get header fontsizes
        if not self.main_body_fontsizes or not self.document_fontsizes:
            return toc
        biggest_body_fontsize = max(self.main_body_fontsizes)
        header_fontsizes = (
            fontsize
            for fontsize in self.document_fontsizes
            if fontsize > biggest_body_fontsize
        )
        header_fontsizes = sorted(header_fontsizes, reverse=True)[:5]
        for block in self.blocks:
            supposed_header_level = None
            if (
                block.orientation != (1.0, 0.0)
                or block.is_empty
                or len(block.text) > 100
            ):
                # do not consider non-horizontal text or long texts as they are unlikely to be headers
                continue
            elif block.fontsize in header_fontsizes:  # likely to be a header
                supposed_header_level = header_fontsizes.index(block.fontsize) + 1
            elif (
                block.fontsize == biggest_body_fontsize
                and block.is_bold
                and not self.main_body_is_bold
            ):  # likely to be a header
                supposed_header_level = len(header_fontsizes) + 1

            if supposed_header_level:
                header_level = (
                    self._get_header_level_using_schema(block.text)
                    or supposed_header_level
                )
                toc_title = TocTitle(
                    text=block.text,
                    level=header_level,
                    page=block.page,
                    source="fontsize",
                    found=True,
                )
                block.section_title = toc_title
                toc.append(toc_title)

        return toc

    def _get_document_main_title(self) -> str:
        """
        Tries to infer the main title of the document.
        It considers the blocks belong to main title if
        they are on first page and their fontsize is bigger
        that the body size of the document.

        Returns :
            (str) : the main title of the document
        """
        spans_on_first_page = [s for s in self.spans if s.page == 0 and not s.is_empty]
        if spans_on_first_page and self.main_body_fontsizes:
            # Get the 2 biggest fontsizes of 1st page
            first_page_biggest_fontsizes = sorted(
                Counter(
                    span.fontsize
                    for span in spans_on_first_page
                    if not span.is_empty and span.orientation == (1.0, 0.0)
                ),
                reverse=True,
            )[:2]
            title_spans = (
                s
                for s in spans_on_first_page
                if s.fontsize not in self.main_body_fontsizes
                and s.fontsize in first_page_biggest_fontsizes
            )
            main_title = " ".join((s.text for s in title_spans)).strip()

            return main_title[:100] + "[...]" if len(main_title) > 100 else main_title

        return ""

    def headers_have_been_found(self, toc: list[TocTitle]) -> bool:
        """Given a potential ToC, determines whether the headers have been
        found in the document.
        If more than half the headers have been found, returns True.
        Otherwise retruns False, indicating that detected the ToC might be wrong.

        Returns:
            bool : True if detected headers have been found in document.
        """
        if len(toc) < 3:
            return False
        headers_found = [header for header in toc if header.found]
        return len(headers_found) / len(toc) > 0.5



import random
from itertools import groupby
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
import pymupdf  # type: ignore : not stubs

from .components import TextBlock, TextLine, TextSpan
from .extract_tables import PdfTable
from .utils import PdfParserState


class PdfPlotter(PdfParserState):
    """Tool that groups methods related to plotting elements obtained by
    the PdfParser. Meant to be as inerited class PdfParser(PdfPlotter)

    Most methods make use of PdfParser's attributes, such as:
    self.spans, self.lines, self.blocks, self.tables and self.document.
    """

    def get_doc_to_draw_on(
        self,
    ) -> pymupdf.Document:
        """Returns a copy of current document in order to draw
        on it without altering the source document"""
        pdfdata = self.document.tobytes()  # type: ignore : missing typing in pymupdf | Document.tobytes(*args) -> bytes
        return pymupdf.open(stream=pdfdata, filetype="pdf")

    def plot_pdf(
        self,
        page_start: int | None = None,
        page_end: int | None = None,
        items_to_plot: list[Literal["span", "line", "block", "table"]] | None = None,
        dpi: int = 100,
    ) -> None:
        """Buils a plot of the pdf file.
        Plots the span's bboxes and tables grid

        Args:
            page_start (int | None): the first page to consider.
            page_end (int | None): the last page to consider.
            items_to_plot (list[str]) : the list of types of elements to plot.
                Can be any subset of ["span", "line", "block", "table"].
            dpi (int): the resolution of the generated image
        """
        page_start = page_start or self.page_start
        page_end = page_end or self.page_end
        items_to_plot = items_to_plot or ["span", "line", "block", "table"]

        doc_to_draw_on = self.get_doc_to_draw_on()

        spans_per_page: dict[int, list[TextSpan]] = PdfPlotter._get_items_per_page(
            self.spans
        )
        lines_per_page: dict[int, list[TextLine]] = PdfPlotter._get_items_per_page(
            self.lines
        )
        blocks_per_page: dict[int, list[TextBlock]] = PdfPlotter._get_items_per_page(
            self.blocks
        )
        tables_per_page: dict[int, list[PdfTable]] = PdfPlotter._get_items_per_page(
            self.tables
        )
        for page in doc_to_draw_on.pages(page_start, page_end):  # type: ignore : missing typing in pymupdf | Document.pages() -> Generator(Page)
            if "span" in items_to_plot and page.number in spans_per_page:  # type: ignore :missing typing in pymupdf Page.number : int
                for span in spans_per_page[page.number]:  # type: ignore :missing typing in pymupdf Page.number : int
                    style = PdfPlotter._get_rect_style(span)
                    page.draw_rect(span.bbox, **style, fill_opacity=0.5)  # type: ignore : missing typing in pymupdf
            if "line" in items_to_plot and page.number in lines_per_page:  # type: ignore :missing typing in pymupdf Page.number : int
                for line in lines_per_page[page.number]:  # type: ignore :missing typing in pymupdf Page.number : int
                    page.draw_rect(line.bbox, color=pymupdf.pdfcolor["black"], width=1)  # type: ignore : missing typing in pymupdf
            if "block" in items_to_plot and page.number in blocks_per_page:  # type: ignore :missing typing in pymupdf Page.number : int
                for block in blocks_per_page[page.number]:  # type: ignore :missing typing in pymupdf Page.number : int
                    page.draw_rect(  # type: ignore : missing typing in pymupdf
                        block.bbox,
                        color=pymupdf.pdfcolor["black"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str, tuple[int, int, int]]
                        width=2,
                        dashes="[3 4] 0",
                    )
            if "table" in items_to_plot and page.number in tables_per_page:  # type: ignore :missing typing in pymupdf Page.number : int
                for table in tables_per_page[page.number]:  # type: ignore :missing typing in pymupdf Page.number : int
                    for cell in table.cells:
                        page.draw_rect(cell, color=pymupdf.pdfcolor["blue"])  # type: ignore : missing typing in pymupdf
                    page.draw_rect(table.bbox, color=pymupdf.pdfcolor["navy"])  # type: ignore : missing typing in pymupdf

            PdfPlotter.show_page(page, dpi=dpi)

    @staticmethod
    def _get_items_per_page(
        items_list: list[TextSpan | TextLine | TextBlock | PdfTable],
    ) -> dict[int, list[TextSpan | TextLine | TextBlock | PdfTable]]:
        """Groups the list of items per page. Takes as input
        any list of objects that have a "page" attribute
        Return a dict of structure is {page_number : sublist_of_items, ...}

        Args:
            items_list (list[TextSpan  |  TextLine  |  TextBlock]): A list
                of items to sort by page.

        Returns:
            dict[int,list]: The map of pages to sublists of items
        """
        return {
            page: list(items_on_page)
            for page, items_on_page in groupby(items_list, key=lambda item: item.page)
        }

    @staticmethod
    def show_page(page: pymupdf.Page, dpi: int = 100) -> None:
        """
        Generates an RGB Pixmap from the page and its rectangles,
        Args:
            page: a Page object of the document with rectangles drawn on it.
            dpi: the resolution of the generated image
        """
        plt.rcParams.update({"font.size": 5})
        pix = page.get_pixmap(dpi=dpi)  # type: ignore : missing typing in pymupdf | Page.get_pixmap(*args) -> Pixmap
        img = np.ndarray([int(pix.h), int(pix.w), 3], dtype=np.uint8, buffer=pix.samples_mv)  # type: ignore : missing typing in pymupdf | Pixmap.h : int ,Pixmap.mv : memoryview
        plt.figure(dpi=dpi)  # type: ignore : missing typing in matplotlib | plt.Figure(*args) -> Figure
        _ = plt.imshow(img, extent=(0, pix.w * 72 / dpi, pix.h * 72 / dpi, 0))  # type: ignore : missing typing in matplotlib | plt.imshow(*args) -> AxesImage

    @staticmethod
    def _get_rect_style(item: TextSpan) -> dict[str, tuple[int, int, int]]:
        """Get the kwargs for styling the rectangle, such as
        edge color and face color, considering the item's attributes

        Args:
            item (TextSpan): the item to get the styling params for

        Returns:
            dict: kwargs to be passed to the Shape.draw_rect() method
        """
        if item.is_header_footer:
            return {
                "color": pymupdf.pdfcolor["red"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
                "fill": pymupdf.pdfcolor["tomato"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
            }
        elif item.isin_table:
            return {
                "color": pymupdf.pdfcolor["steelblue"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
                "fill": pymupdf.pdfcolor["lightblue"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
            }
        else:
            return {
                "color": pymupdf.pdfcolor["goldenrod"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
                "fill": pymupdf.pdfcolor["gold"],  # type: ignore : missing typing in pymupdf | pymupdf.pdfcolor : dict[str | tuple[int, int, int]]
            }

    def plot_drawings(
        self,
        items_to_draw: list[Literal["l", "c", "re", "qu", "p"]] | None = None,
        page_start: int | None = None,
        page_end: int | None = None,
        dpi: int = 100,
    ):
        """Draws the raw drawings extracted from the page.
        Mainly used for debug.

        Args:
            items_to_draw (list[Literal["l", "c", "re", "qu"]], optional): the elements to draw.
                See pymupdf.get_drawing() documentation for more info.
                Defaults to None meaning all elements are drawn.
            page_start (int | None): the first page to consider.
            page_end (int | None): the last page to consider.
            dpi (int) : the resolution. Defaults to 100.
        """
        items_to_draw = items_to_draw or ["l", "c", "re", "qu", "p"]
        doc_to_draw_on = self.get_doc_to_draw_on()
        page_start = page_start or self.page_start
        page_end = page_end or self.page_end

        for page in doc_to_draw_on.pages(page_start, page_end):  # type: ignore : missing typing in pymupdf | Document.pages() -> Generator(Page)
            drawings: list[dict[str, tuple[str, tuple[Any]]]] = page.get_drawings()  # type: ignore : missing typing in pymupdf | Page.get_drawings() -> list[dict[Any]]
            for i, drawing in enumerate(drawings):
                for j, item in enumerate(drawing["items"]):
                    if not item[0] in items_to_draw:
                        continue
                    match item[0]:
                        case "l":
                            page.draw_line(*item[1:], color=(1, 0, 0.7))  # type: ignore : missing typing in pymupdf
                            page.insert_text((item[1].x, item[1].y), f"d{i}i{j}l")  # type: ignore : missing typing in pymupdf | Point.x/y -> float
                        case "c":
                            page.draw_bezier(*item[1:], color=(1, 1, 0))  # type: ignore : missing typing in pymupdf | Document.pages() -> Generator(Page)
                        case "re":
                            page.draw_rect(  # type: ignore : missing typing in pymupdf | Document.pages() -> Generator(Page)
                                item[1],
                                color=(
                                    random.random(),
                                    random.random(),
                                    random.random(),
                                ),
                            )
                            page.insert_text((item[1].x0, item[1].y0), f"d{i}i{j}r")  # type: ignore : missing typing in pymupdf
                        case "qu":
                            page.draw_quad(item[1], color=(0, 0, 1))  # type: ignore : missing typing in pymupdf
                        case "p":
                            page.draw_circle(item, 3)  # type: ignore : missing typing in pymupdf

            PdfPlotter.show_page(page, dpi=dpi)

    def plot_parsed_tables(
        self, page_start: int | None = None, page_end: int | None = None, dpi: int = 100
    ):
        """Plots a parsed table"""
        if not self.tables:
            raise ValueError(
                "No tables to be plotted. Make sure a document has been parsed."
            )
        page_start = page_start or self.page_start
        page_end = page_end or self.page_end
        doc_to_draw_on = self.get_doc_to_draw_on()
        for page in doc_to_draw_on.pages(page_start, page_end):  # type: ignore : missing typing in pymupdf | Document.pages() -> Generator(Page)
            tables = self.table_finder.build_tables(page)
            if not tables:
                continue
            for lines, intersections, cells in tables:
                color = (
                    random.uniform(0.4, 0.7),
                    random.uniform(0.4, 0.7),
                    random.uniform(0.4, 0.7),
                )
                page = PdfPlotter.draw_lines(
                    lines, page, color=[c + random.uniform(-0.1, 0.1) for c in color]
                )
                page = PdfPlotter.draw_rectangles(
                    cells, page, color=[c - 0.4 for c in color]
                )
                page = PdfPlotter.draw_points(
                    intersections, page, color=[c + 0.2 for c in color]
                )
            PdfPlotter.show_page(page, dpi=dpi)

    @staticmethod
    def draw_points(
        points_coord: list[tuple[float, float]],
        page_to_draw_on: pymupdf.Page,
        color: tuple[float, float, float] = (0, 0, 0),
    ) -> pymupdf.Page:
        """Draws a list of points on provided page"""
        for point in points_coord:
            page_to_draw_on.draw_circle(pymupdf.Point(point), 3, fill=color)  # type: ignore : missing typing in pymupdf | Point.x/y -> float

        return page_to_draw_on

    @staticmethod
    def draw_rectangles(
        rectangle_coords: list[pymupdf.Rect],
        page_to_draw_on: pymupdf.Page,
        color: tuple[float, float, float] = (0, 0, 0),
    ) -> pymupdf.Page:
        """Draws a list of rectangle on provided page"""
        for rect in rectangle_coords:
            page_to_draw_on.draw_rect(list(rect), color=color, width=2)  # type: ignore : missing typing in pymupdf | Point.x/y -> float

        return page_to_draw_on

    @staticmethod
    def draw_lines(
        lines: list[tuple[float, float, float, float]],
        page_to_draw_on: pymupdf.Page,
        color: tuple[float, float, float] = (0, 0, 0),
    ) -> pymupdf.Page:
        """Draws a list of lines on provided page"""
        for line in lines:
            page_to_draw_on.draw_line(  # type: ignore : missing typing in pymupdf | Point.x/y -> float
                (line[0], line[1]), (line[2], line[3]), color=color, width=6
            )

        return page_to_draw_on

    def plot_reading_order(
        self, page_start: int | None = None, page_end: int | None = None, dpi: int = 100
    ):
        """Plots the detected reading order of the blocks"""
        page_start = page_start or self.page_start
        page_end = page_end or self.page_end
        doc_to_draw_on = self.get_doc_to_draw_on()

        blocks_per_page: dict[int, list[TextBlock]] = PdfPlotter._get_items_per_page(
            self.blocks
        )
        for page in doc_to_draw_on.pages(page_start, page_end):  # type: ignore : missing typing in pymupdf | Document.pages() -> Generator(Page)
            if not page.number in blocks_per_page:  # type: ignore : missing typing in pymupdf | Page.number -> int
                continue

            prev_block_location = (0, 0)
            for block in blocks_per_page[page.number]:
                block_location = (
                    block.bbox.x1 - (block.bbox.x1 - block.bbox.x0) / 2,  # type: ignore : missing typing in pymupdf
                    block.bbox.y1 - (block.bbox.y1 - block.bbox.y0) / 2,  # type: ignore : missing typing in pymupdf
                )
                page.draw_rect(list(block.bbox), fill=(0.6, 0.6, 0.8), width=2)  # type: ignore : missing typing in pymupdf
                page.insert_text(block_location, f"{block.order}", fontsize=30)  # type: ignore : missing typing in pymupdf
                page.draw_line(prev_block_location, block_location, color=(0.8, 0.3, 0.3), width=4)  # type: ignore : missing typing in pymupdf
                page.draw_circle(block_location, 6, fill=(0.5, 0.2, 0.2), color=(0.1, 0.1, 0.1))  # type: ignore : missing typing in pymupdf
                prev_block_location = block_location

            PdfPlotter.show_page(page, dpi=dpi)



import re
from io import BytesIO
from pathlib import Path
from typing import Literal

import pandas as pd

from ...core.components import MarkdownDoc
from ..abstract_parser import AbstractParser


class ExcelParser(AbstractParser):
    """Parser for spreadsheets, such as Excel workbooks (.xslx). For a list of handled filetypes,
    refer to https://pandas.pydata.org/docs/reference/api/pandas.read_excel.html"""

    def __init__(
        self,
        output_format: Literal["markdown_table", "json_lines", "auto"] = "auto",
    ) -> None:
        """Initializes an Excel parser

        Args:
            output_format (Literal["markdown_table", "json_lines", "auto], optional): the output format of the parsed document.
                - markdown_table : uses tabula to build a markdown-formatted table.
                - json_lines : each row of the table will be output as a JSON line. Better for chunking as headers are preserved.
                - auto : will detect which format is the more suitable. CSV-like sheet will be converset to JSON lines.
                Defaults to "auto".
        """
        self.output_format = output_format

    def parse_file(self, filepath: str) -> MarkdownDoc:
        """Parses a excel-like file to markdown.

        Args:
            filepath (str): the path to the excel-like file.

        Returns:
            MarkdownDoc: the markdown formatted excel file.
        """
        sheets = self.read_file(filepath)
        md_string = self.convert_sheets_to_output_format(sheets)

        return MarkdownDoc.from_string(md_string)

    def parse_string(self, string: bytes) -> MarkdownDoc:
        """Parses a bytes string representing an excel file.

        Args:
            string (bytes): the excel as a byte string.

        Returns:
            MarkdownDoc: the markdown formatted excel file
        """
        sheets = pd.read_excel(BytesIO(string), sheet_name=None)  # type: ignore | missing typing in pandas.
        md_string = self.convert_sheets_to_output_format(sheets)

        return MarkdownDoc.from_string(md_string)

    def read_file(self, filepath: str) -> dict[str, pd.DataFrame]:
        """Read the provided filepath.

        Args:
            filepath (str): path to the file.

        Returns:
            dict[str, pd.DataFrame]: a mapping containing {sheet_name: corresponding_dataframe.}
        """
        path = Path(filepath)
        if path.suffix.lower() not in [
            ".xls",
            ".xlsx",
            ".xlsm",
            ".xlsb",
            ".odf",
            ".ods",
            ".odt",
        ]:
            raise ValueError(f"ExcelParser cannot parse {path.suffix.lower()} files.")
        sheets = pd.read_excel(filepath, sheet_name=None)  # type: ignore | missing typing in pandas.

        return sheets

    def convert_sheets_to_output_format(self, sheets: dict[str, pd.DataFrame]) -> str:
        """Handle the conversion of the sheets obtained from
        pandas.read_excel() method to the specified output format.

        Args:
            sheets (dict[str, pd.DataFrame]): the sheets returned from pd.read_excel(sheet_name=None).

        Returns:
            str: the formatted string
        """
        output_string = ""

        for sheet_name, df in sheets.items():
            output_string += f"## {sheet_name}\n\n"

            format_to_use = self.output_format
            if self.output_format == "auto":
                format_to_use = self._determine_best_format(df)

            match format_to_use:
                case "markdown_table":
                    output_string += self.convert_df_to_markdown_table(df) + "\n\n"
                case "json_lines":
                    output_string += self.convert_df_to_json_lines(df) + "\n\n"
                case _:
                    raise ValueError(
                        f"Invalid value for argument 'output_format': expected one of ['markdown_table', 'json_lines']. Got '{format_to_use}'."
                    )

        return output_string

    @staticmethod
    def convert_df_to_markdown_table(df: pd.DataFrame) -> str:
        """Converts a DataFrame to markdown.
        Wraps tabula's method pd.DataFrame.to_markdown()
        between pre and post processing.
        Preprocess :
        - Remove \n in text columns
        PostProcess :
        - Replace multiple spaces with 2 spaces.

        Args:
            df (pd.DataFrame): the dataframe to convert.

        Returns:
            str: a markdown formatted table.
        """
        dtypes = df.apply(  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
            lambda x: pd.api.types.infer_dtype(x, skipna=True)  # type: ignore | x: pd.Series[Any/co]
        )
        string_cols = dtypes[dtypes == "string"].index  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
        df[string_cols] = df[string_cols].apply(lambda x: x.str.replace("\n", " "))  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
        md_string = df.to_markdown(index=False)
        md_string = re.sub(r"\s{3,}", "  ", md_string)
        md_string = re.sub(r"-{3,}", "---", md_string)

        return md_string

    @staticmethod
    def convert_df_to_json_lines(df: pd.DataFrame) -> str:
        """Converts a DataFrame to json lines.

        Args:
            df (pd.DataFrame): the dataframe to convert.

        Returns:
            str: the json lines.
        """
        return df.to_json(orient="records", force_ascii=False, lines=True)  # type: ignore | df.to_json() -> str

    def _determine_best_format(
        self, df: pd.DataFrame
    ) -> Literal["markdown_table", "json_lines"]:
        """Determine the best output format based on DataFrame characteristics.

        CSV-like sheets (good for json_lines):
        - Have clear column headers
        - Data is organized in rows with consistent structure
        - Most cells contain data (not sparse)

        Dashboard-like sheets (better for markdown_table):
        - Have merged cells or complex layout
        - Sparse data with many empty cells
        - Mixed content types in irregular patterns

        Args:
            df (pd.DataFrame): the dataframe to analyze.

        Returns:
            Literal["markdown_table", "json_lines"]: the recommended format
        """
        # Check if sheet looks CSV-like
        if self._is_csv_like(df):
            return "json_lines"
        else:
            return "markdown_table"

    @staticmethod
    def _is_csv_like(df: pd.DataFrame) -> bool:
        """Check if a DataFrame has CSV-like characteristics.

        A sheet is CSV-like if it has:
        - High data density (>= 60%)
        - Good header quality (>= 70%)
        - Consistent row structure (>= 70%)

        Args:
            df (pd.DataFrame): the dataframe to analyze.

        Returns:
            bool: True if the DataFrame appears to be CSV-like
        """
        if df.empty or len(df) < 2:
            return False

        # Check data density (percentage of non-null cells)
        total_cells = df.size
        non_null_cells = df.count().sum()
        data_density = non_null_cells / total_cells if total_cells > 0 else 0

        # Check column header quality (non-null, non-empty column names)
        valid_headers = sum(
            1
            for col in df.columns
            if col and str(col).strip() and not str(col).startswith("Unnamed")
        )
        header_quality = valid_headers / len(df.columns) if len(df.columns) > 0 else 0

        # Check row consistency (similar number of non-null values per row)
        row_counts = df.count(axis=1)
        if len(row_counts) > 1:
            row_consistency = (
                1 - (row_counts.std() / row_counts.mean())
                if row_counts.mean() > 0
                else 0
            )
        else:
            row_consistency = 1.0

        return data_density >= 0.6 and header_quality >= 0.8 and row_consistency >= 0.7



import os
import re
from collections import defaultdict
from operator import attrgetter

from ....core.components import MarkdownDoc, MarkdownLine
from .utils import PdfParserState


class PdfExport(PdfParserState):
    """This class is intended to be inherited by the PdfParser.
    It groups all functions related to exporting the Pdf based
    on the elements parsed by the PdfParser. Uses attributes
    such as self.spans, self.lines and self.blocks
    """

    def to_markdown(
        self,
        keep_track_of_page: bool = False,
    ) -> str | dict[int, str]:
        """Export parsed content to markdown.

        Args:
            keep_track_of_page (bool) : if True, returns a dict with
                dict[page_number : string]. Page number is 0-based
                and page 0 is reserved for main document title.
                If False returns a string for the whole document.
                Defaults to False.

        Returns:
            str | dict[int, str]: the markdown formatted string, split by page if keep_track_of_page = True.
        """
        prefix = f"# {self.main_title}\n\n" if self.main_title else ""
        string_per_page = defaultdict(str)
        items_to_export = sorted(self.blocks + self.tables, key=attrgetter("order"))
        for item in items_to_export:
            if not item.is_header_footer:
                string_per_page[item.page] += "\n\n" + item.to_markdown()
        string_per_page = {
            page: PdfExport._cleanup_md_string(md_string)
            for page, md_string in string_per_page.items()
        }

        if keep_track_of_page:
            return string_per_page

        return prefix + "\n\n".join(string_per_page.values())

    def to_markdown_doc(self) -> MarkdownDoc:
        """Generated the markdown doc to be fed in the
        MarkdownChunker.

        Returns:
            MarkdownDoc: the formatted markdown doc
        """
        items_to_export = sorted(self.blocks + self.tables, key=attrgetter("order"))
        md_lines: list[MarkdownLine] = []
        line_idx_counter = 0

        main_title = f"# {self.main_title}\n\n" if self.main_title else ""
        md_lines.append(MarkdownLine(text=main_title, line_idx=-1, page=0))

        for item in items_to_export:
            if item.is_header_footer:
                continue
            line_text = (
                "\n\n" + PdfExport._cleanup_md_string(item.to_markdown()) + "\n\n"
            )
            line_count = len(line_text.split("\n"))
            md_lines.append(
                MarkdownLine(
                    text=line_text,
                    line_idx=line_count,
                    isin_code_block=False,
                    page=item.page,
                )
            )
            line_idx_counter += line_count

        return MarkdownDoc(content=md_lines)

    @staticmethod
    def _cleanup_md_string(md_string: str) -> str:
        """Performs various cleanup operation
        of the markwdown string output by to_markdown()

        Args:
            md_string (str): the markdown string

        Returns:
            str : the cleaned up string
        """
        md_string = re.sub(r"\*\* *\*\*", " ", md_string)
        md_string = md_string.replace(" ** ", "**\n")
        md_string = re.sub(r"\n{3,}", "\n\n", md_string)

        return md_string.strip()

    def save_markdown(self, output_filepath: str = "./output.md") -> None:
        """Generates the markdown export of the pdf
        and saves it in a file

        Args:
            output_filpath (str): the file path where the md is saved
        """
        md_string = self.to_markdown()
        if not os.path.exists(os.path.dirname(output_filepath)):
            os.makedirs(os.path.dirname(output_filepath))
        with open(output_filepath, "w", encoding="utf-8") as f:
            f.write(md_string)



import re
from operator import attrgetter
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
import pymupdf  # type: ignore | No stubs

from .components import TextSpan
from ....core.logger import LOGGER


class Cell(pymupdf.Rect):
    """A cell object. It inherit
    the pymupdf.Rect class but has PdfParser's spans
    bound to it
    """

    spans: list[TextSpan]

    def __init__(self, x0: float, y0: float, x1: float, y1: float) -> None:
        super().__init__(x0, y0, x1, y1)  # type: ignore : missing typing in pymuPdf : Rect.x0, Rect.y0, Rect.x1, Rect.y1 are all float
        # State
        self.spans = []


class PdfTable:
    """This class is intended to be used to parse the table of the document
    Note that pyMuPdf already implements a "Table" class.
    Though it has several issues :
    - wrongly parsed cell text (artifacts)
    - bad handling of merged cells
    - not fitting well with the mechanic of the PdfParser
    """

    cells: list[Cell]
    page: int

    def __init__(self, cells: list[Cell], page: int) -> None:
        """Initialize Table

        Args:
            cells (list[Cell]): a list of Cell object that are part of the table
            page (int): the page number the table belongs to
        """
        self.cells = cells
        self.page = page

    @property
    def bbox(self) -> pymupdf.Rect:
        return pymupdf.Rect(
            min(self.cells, key=lambda cell: cell.x0).x0,  # type: ignore : missing typing in pymuPdf | Rect.x0 : float
            min(self.cells, key=lambda cell: cell.y0).y0,  # type: ignore : missing typing in pymuPdf | Rect.y0 : float
            max(self.cells, key=lambda cell: cell.x1).x1,  # type: ignore : missing typing in pymuPdf | Rect.x1 : float
            max(self.cells, key=lambda cell: cell.y1).y1,  # type: ignore : missing typing in pymuPdf | Rect.y1 : float
        )

    @property
    def is_header_footer(self) -> bool:
        return all(span.is_header_footer for cell in self.cells for span in cell.spans)

    @property
    def order(self) -> int:
        return min(
            (span for cell in self.cells for span in cell.spans),
            key=attrgetter("order"),
        ).order

    def get_table_grid(self) -> list[list[Cell]]:
        """Gets the grid of the table.
        As some self.cells might be merged cells, this
        method get the grid of cells as if no cells were merged

        Returns:
            list[list[Cells]] : the list of cells, organized by row
        """
        y_lines = sorted(
            list(
                set(cell.y0 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.y0 : float
                | set(cell.y1 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.y1 : float
            )
        )
        x_lines = sorted(
            list(
                set(cell.x0 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.x0 : float
                | set(cell.x1 for cell in self.cells)  # type: ignore : missing typing in pymuPdf | Rect.x1 : float
            )
        )
        grid_cells: list[list[Cell]] = []
        for y0, y1 in zip(y_lines[:-1], y_lines[1:]):
            row_cells: list[Cell] = []
            for x0, x1 in zip(x_lines[:-1], x_lines[1:]):
                row_cells.append(Cell(x0, y0, x1, y1))
            grid_cells.append(row_cells)

        return grid_cells

    def to_pandas(self) -> pd.DataFrame:
        """Transforms this table to a pd.DataFrame

        Returns:
            pd.DataFrame : the table as dataframe
        """
        grid_cells = self.get_table_grid()
        df_constructor: list[list[str]] = []
        for grid_row in grid_cells:
            row_text: list[str] = []
            for grid_cell in grid_row:
                cell_text = ""
                for cell in self.cells:
                    if cell.contains(grid_cell):  # type: ignore : missing typing in pymuPdf | Rect.contains(r: Rect | Point) -> bool
                        for span in cell.spans:
                            cell_text += (
                                " " + span.text
                                if not span.link
                                else f"[{span.text}]({span.link.uri})"
                            )
                row_text.append(cell_text)
            df_constructor.append(row_text)

        df = (
            pd.DataFrame(df_constructor[1:], columns=df_constructor[0])
            .drop_duplicates()
            .dropna(axis=0, how="all")  # type: ignore | type of dropna in partially unknown
            .dropna(axis=1, how="all")  # type: ignore | type of dropna in partially unknown
        )

        return df

    def to_markdown(self) -> str:
        """Builds a markdown string from the provided table

        Returns:
            str: the markdown string
        """
        table_as_md = self.to_pandas().to_markdown(index=False)
        table_as_md = re.sub(r"\s{3,}", "  ", table_as_md)
        table_as_md = re.sub(r"-{3,}", "---", table_as_md)

        return table_as_md

    def __str__(self) -> str:
        return self.to_markdown()


class TableFinder:
    """Superseeds pymupdf's TableFinder class. Aims at being faster
    while maintaining table parsing capabilities
    """

    def __init__(self, snap_tolerance: int = 3, line_width_threshold: int = 5):
        """Init a tablefinder

        Args:
            snap_tolerance (int, optional): the tolerance threshold (in point) to consider
                two lines are intersecting or aligned. Defaults to 3.
            line_width_threshold (int, optional): natively, most lines come as a rectangle
                in the pdf, to take into account the width of the line.
                This arg specifies a threshold : all rectangle having a width below it is considered a line.
                Defaults to 5.
        """
        self.snap_tolerance = snap_tolerance
        self.line_width_threshold = line_width_threshold

    def build_tables(
        self, page: pymupdf.Page
    ) -> list[
        tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]
    ]:
        """Method that wraps all the logic
        to parse the tables on a page of a pdf document.

        Args:
            page (pymupdf.Page): a page of a document

        Returns:
            list[tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]]:
                Each item of the list is a tuple that contains 3 numpy arrays:
                    - coordinates of detected lines, where each row is (x1, y1, x2, y2). Shape : (n_lines, 4).
                    - coordinates of intersections between lines, where each row is (x, y). Shape : (n_intersections, 2).
                    - coordinates of cells, where each row is coordinates of rectangles as where each row is (x1, y1, x2, y2). Shape : (n_cells, 2).
        """
        lines_coordinates = self._get_table_lines(page)
        if not lines_coordinates.size:
            return []
        lines_grouped_by_table = self.group_lines_by_table(lines_coordinates)
        # tuples of (line_coord, intersections and cells) for each table. We might only need the cells later if we do not need plotting
        parsed_tables: list[
            tuple[
                npt.NDArray[np.float32],
                npt.NDArray[np.float32],
                npt.NDArray[np.float32],
            ]
        ] = []
        for table_lines in lines_grouped_by_table:
            try:
                parsed_tables.append(self.build_table(table_lines))
            except AssertionError:
                LOGGER.warning(
                    "Table parsing failed on page %i (likely because detected lines are not a table).",
                    page.number,  # type: ignore | missing typing in pymupdf page.number -> int
                )
                pass
        # remove tables with no cells
        return [tab for tab in parsed_tables if tab[2].size]

    def _get_table_lines(self, page: pymupdf.Page) -> npt.NDArray[np.float32]:
        """Gets the lines that are likely to belong to a table. Proceed like so:
        - Get the drawings on the page (vectors)
        - Removes the drawings that are due to annotations (i.e. rectangles used to highlight text)
        - Convert rectangles coordinates to 4 lines coordinates
        - Remove very short lines that are likely to belong to vector designs

        Args:
            page (pymupdf.Page): the page to get the drawings from.

        Returns:
            npt.NDArray[np.float32]: an array of line coordinates of shape (n_lines, 4)
                where each row is x1, y1, x2, y2.
        """
        drawings = page.get_drawings()  # type: ignore missing typing in pymupdf
        drawings = TableFinder._remove_drawings_from_annotations(drawings, page)
        drawings = self._convert_rectangles_to_lines(drawings)
        # Get coordinates of all lines
        line_coordinates = np.array(
            [
                (item[1].x, item[1].y, item[2].x, item[2].y)
                for drawing in drawings
                for item in drawing["items"]
                if item[0] == "l"
            ]
        ).round()
        if not line_coordinates.size:
            return np.empty(shape=(0, 4))
        line_coordinates = TableFinder._filter_lines(line_coordinates)

        return line_coordinates

    @staticmethod
    def _remove_drawings_from_annotations(
        drawings: list[dict[str, Any]], page: pymupdf.Page
    ) -> list[dict[str, Any]]:
        """Removes the drawings that are likely to be due to annotations, i.e rectangles
        used to highlight text.

        Args:
            drawings (list[dict[Any]]): the drawings obtained from Page.get_drawings()
            page (pymupdf.Page): the page to get the drawings from.

        Returns:
            list[dict[Any]]: the drawings that are not due to annotations.
        """
        annotations = list(page.annots())  # type: ignore :: missing typing in pymupdf -> Page.annots() -> Generator[Annots]
        if not annotations:
            return drawings

        filtered_drawings: list[dict[str, Any]] = [
            drawing
            for drawing in drawings
            if not any(annotation.rect.contains(drawing["rect"]) for annotation in annotations)  # type: ignore :: missing typing in pymupdf -> Rect.contains() -> bool
        ]

        return filtered_drawings

    def _convert_rectangles_to_lines(
        self, drawings: list[tuple[Any]]
    ) -> list[dict[str, list[tuple[Any]]]]:
        """Converts the rectangles to lines if their width
        is below self.line_width_threshold.

        Args:
            drawings (list[tuple[Any]]): the drawings extracted from a pymupdf.Page.

        Returns:
            list[dict[str, list[tuple[Any]]]]: the drawings, same format as pumupdf.Page.get_cdrawings()'s return,
                with some rectangles converted to lines.
        """
        processed_drawings: list[dict[str, list[tuple[Any]]]] = []
        for drawing in drawings:
            drawing_items: list[tuple[Any]] = []
            for item in drawing["items"]:
                if item[0] == "re":
                    if item[1].width < self.line_width_threshold:  # type: ignore : missing typing in pymupdf rect.witdh : float
                        mid_x = (item[1].x1 - item[1].x0) / 2 + item[1].x0  # type: ignore : missing typing in pymupdf rect.x0/x1 : float
                        item = (
                            "l",
                            pymupdf.Point(mid_x, item[1].y0),  # type: ignore : missing typing in pymupdf rect.y0/y1 : float
                            pymupdf.Point(mid_x, item[1].y1),  # type: ignore : missing typing in pymupdf rect.y0/y1 : float
                        )
                    elif item[1].height < self.line_width_threshold:  # type: ignore : missing typing in pymupdf rect.height : float
                        mid_y = (item[1].y1 - item[1].y0) / 2 + item[1].y0  # type: ignore : missing typing in pymupdf rect.y0/y1 : float
                        item = (
                            "l",
                            pymupdf.Point(item[1].x0, mid_y),  # type: ignore : missing typing in pymupdf rect.x0/x1 : float
                            pymupdf.Point(item[1].x1, mid_y),  # type: ignore : missing typing in pymupdf rect.x0/x1 : float
                        )
                drawing_items.append(item)
            processed_drawings.append({"items": drawing_items})

        return processed_drawings

    @staticmethod
    def _filter_lines(
        line_coordinates: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """filter the detected lines if they might belong to tables"""
        # only keep vertical or horizontal lines
        line_coordinates = line_coordinates[
            (line_coordinates[:, 0] == line_coordinates[:, 2])
            | ((line_coordinates[:, 1] == line_coordinates[:, 3]))
        ]
        # only keep lines of minimum size 5 points
        # TODO : Determine if threshold of 5 is good or not. Maybe find dynamic threshold based on page size.
        line_coordinates = line_coordinates[
            (np.abs(line_coordinates[:, 0] - line_coordinates[:, 2]) > 5)
            | (np.abs(line_coordinates[:, 1] - line_coordinates[:, 3]) > 5)
        ]

        return line_coordinates

    def build_table(
        self, lines_coordinates: npt.NDArray[np.float32]
    ) -> tuple[
        npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]
    ]:
        """Considering lines coordinate of 1 tables,
        returns the cells of that table

        Args:
            lines_coordinates (npt.NDArray[np.float32]): Array of shape (n lines, 4) representing
                the coordinates of the lines with each row being (x0, y0, x1, y1) coordinates.

        Returns:
            tuple[npt.NDArray[np.float32], npt.NDArray[np.float32], npt.NDArray[np.float32]]]:
                Tuple that contains 3 numpy arrays:
                    - coordinates of detected lines, where each row is (x1, y1, x2, y2). Shape : (n_lines, 4).
                    - coordinates of intersections between lines, where each row is (x, y). Shape : (n_intersections, 2).
                    - coordinates of cells, where each row is coordinates of rectangles as where each row is (x1, y1, x2, y2). Shape : (n_cells, 2).
        """
        # Get intersections of lines
        intersections = self.get_line_intersections(lines_coordinates)
        if not intersections.size:
            return lines_coordinates, np.empty((0, 2)), np.empty((0, 4))
        intersections = self.normalize_table_grid(intersections)

        lines_coordinates = self._normalize_lines_grid(lines_coordinates, intersections)
        lines_coordinates = TableFinder.subdivide_lines(
            lines_coordinates, intersections
        )
        # Get cells
        cells = self._get_cells(intersections, lines_coordinates)
        if not cells.size or not TableFinder._table_sanity_check(cells):
            return lines_coordinates, intersections, np.empty((0, 4))

        return lines_coordinates, intersections, cells

    def get_line_intersections(
        self, coordinates: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """
        Find all intersection points between line segments.

        Args:
            coordinates (npt.NDArray[np.float32]): A numpy array of shape (n_lines, 4) where each row is the coordinates
                of the points of the lines like (x0, y0, x1, y1).

        Returns:
            (npt.NDArray[np.float32]): A numpy array of shape (n_intersections, 2) where each row is a the coordinates
                an intersection point between 2 segments like (x, y).
        """
        # Extract coordinates
        x1, y1, x2, y2 = self._bounding_boxes(coordinates)
        # Create matrices for pairwise combinations
        n = coordinates.shape[0]
        x1_tile = np.tile(x1, (n, 1))
        y1_tile = np.tile(y1, (n, 1))
        x2_tile = np.tile(x2, (n, 1))
        y2_tile = np.tile(y2, (n, 1))
        x3_tile = x1_tile.T
        y3_tile = y1_tile.T
        x4_tile = x2_tile.T
        y4_tile = y2_tile.T
        # Calculate denominators
        denom = (x1_tile - x2_tile) * (y3_tile - y4_tile) - (y1_tile - y2_tile) * (
            x3_tile - x4_tile
        )
        # Avoid division by zero for parallel lines
        denom[denom == 0] = np.nan
        # Calculate intersection points
        px = (
            (x1_tile * y2_tile - y1_tile * x2_tile) * (x3_tile - x4_tile)
            - (x1_tile - x2_tile) * (x3_tile * y4_tile - y3_tile * x4_tile)
        ) / denom
        py = (
            (x1_tile * y2_tile - y1_tile * x2_tile) * (y3_tile - y4_tile)
            - (y1_tile - y2_tile) * (x3_tile * y4_tile - y3_tile * x4_tile)
        ) / denom
        # Check if intersection points are within the segments
        within_segment1 = (
            (np.minimum(x1_tile, x2_tile) - self.snap_tolerance <= px)
            & (px <= np.maximum(x1_tile, x2_tile) + self.snap_tolerance)
            & (np.minimum(y1_tile, y2_tile) - self.snap_tolerance <= py)
            & (py <= np.maximum(y1_tile, y2_tile) + self.snap_tolerance)
        )
        within_segment2 = (
            (np.minimum(x3_tile, x4_tile) - self.snap_tolerance <= px)
            & (px <= np.maximum(x3_tile, x4_tile) + self.snap_tolerance)
            & (np.minimum(y3_tile, y4_tile) - self.snap_tolerance <= py)
            & (py <= np.maximum(y3_tile, y4_tile) + self.snap_tolerance)
        )
        valid_intersections = within_segment1 & within_segment2
        # Extract valid intersection points
        intersections = np.column_stack(
            (px[valid_intersections], py[valid_intersections])
        )

        return np.unique(intersections, axis=0)

    def normalize_table_grid(
        self, grid_points: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Normalizes the grid determined by intersections of lines.
        It aligns points vertically and horizontally
        if they are almost on the same x and y axis,
        and merges the points that are close together.

        Args:
            grid_points (npt.NDArray): the (x, y) coordinates of points defining the table grid,
                i.e intersections point of lines. As an array of shape (n_points, 2).

        Returns:
            npt.NDArray[np.float32] : the aligned points.
        """
        x = grid_points[:, 0]
        groups_by_x = self._get_grouped_idxes(x)
        for group in groups_by_x:
            grid_points[group, 0] = np.mean(grid_points[group, 0])

        y = grid_points[:, 1]
        groups_by_y = self._get_grouped_idxes(y)
        for group in groups_by_y:
            grid_points[group, 1] = np.mean(grid_points[group, 1])

        return np.unique(grid_points, axis=0)

    def _get_grouped_idxes(self, values: list[float]) -> list[npt.NDArray[np.float32]]:
        """Considering a list of values, groups the values together if
        they are closer than the specified threshold, and returns the indices
        of the groups.
        For examples, if values = [2, 3, 5, 10] and threshold = 2,
        it will return [[0,1], [2], [3]]"""
        x = np.sort(values)
        diff = x[1:] - x[:-1]
        gps = np.concatenate([[0], np.cumsum(diff >= self.line_width_threshold)])

        return [np.argsort(values)[gps == i] for i in range(gps[-1] + 1)]

    def _normalize_lines_grid(
        self,
        line_coords: npt.NDArray[np.float32],
        intersections: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """Remaps the points of all the detected lines on the detected intersections.

        Args:
            line_coords (npt.NDArray[np.float32]): the coordinates of the detected lines,
                as an array of shape (n_lines, 4)
            intersections (npt.NDArray[np.float32]) : the coordinates of the intersections,
                as an array of shape (n_intersections, 2)

        Returns:
            npt.NDArray[np.float32]: _description_
        """
        sorted_x_y = np.sort(intersections, axis=0)
        x_targets = np.unique(sorted_x_y[:, 0])
        y_targets = np.unique(sorted_x_y[:, 1])

        x_bins = (x_targets[1:] + x_targets[:-1]) / 2
        y_bins = (y_targets[1:] + y_targets[:-1]) / 2

        lines_as_points = line_coords.reshape(-1, 2)
        normalized_x = x_targets[np.digitize(lines_as_points[:, 0], x_bins)]
        normalized_y = y_targets[np.digitize(lines_as_points[:, 1], y_bins)]
        points_as_lines = np.stack([normalized_x, normalized_y], axis=-1).reshape(-1, 4)
        # reorder xs and ys so that (x0, y0) is topleft and (x1, y1) is bottomright
        points_as_lines = np.column_stack(TableFinder._bounding_boxes(points_as_lines))

        return points_as_lines

    @staticmethod
    def _bounding_boxes(
        coordinates: npt.NDArray[np.float32],
    ) -> tuple[
        npt.NDArray[np.float32],
        npt.NDArray[np.float32],
        npt.NDArray[np.float32],
        npt.NDArray[np.float32],
    ]:
        """Calculate bounding boxes for line segments."""
        x1 = np.minimum(coordinates[:, 0], coordinates[:, 2])
        y1 = np.minimum(coordinates[:, 1], coordinates[:, 3])
        x2 = np.maximum(coordinates[:, 0], coordinates[:, 2])
        y2 = np.maximum(coordinates[:, 1], coordinates[:, 3])

        return x1, y1, x2, y2

    def _bboxes_intersect_or_nearly(
        self,
        bbox_a: tuple[float],
        bbox_b: tuple[float],
    ) -> bool:
        """Check if two bounding boxes of lines a and b intersect or are nearly intersecting.
        Arguments must be bbox coordinates as a tuple representing (x1, y1, x2, y2)."""
        x1a, y1a, x2a, y2a = bbox_a
        x1b, y1b, x2b, y2b = bbox_b
        return (
            (x1a <= x2b + self.snap_tolerance)
            & (x2a + self.snap_tolerance >= x1b)
            & (y1a <= y2b + self.snap_tolerance)
            & (y2a + self.snap_tolerance >= y1b)
        )

    def _build_adjacency_matrix(
        self, coordinates: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.bool_]:
        """Builds the adjencency matrix, i.e matrix denoting which
        lines intersect with one another.

        Args:
            coordinates (npt.NDArray[np.float32]): a 2D array of shape (n_lines, 4) representing line coordinates.

        Returns:
            npt.NDArray[np.bool_]: A 2D matrix of shape (n_lines, n_lines) representing which
                line intersects with other lines.
        """
        x1, y1, x2, y2 = TableFinder._bounding_boxes(coordinates)

        n = len(coordinates)
        adjacency_matrix = np.zeros((n, n), dtype=bool)
        for i in range(n):
            for j in range(i + 1, n):
                if self._bboxes_intersect_or_nearly(
                    (x1[i], y1[i], x2[i], y2[i]), (x1[j], y1[j], x2[j], y2[j])
                ):
                    adjacency_matrix[i, j] = True
                    adjacency_matrix[j, i] = True

        return adjacency_matrix

    @staticmethod
    def _connected_components(
        adjacency_matrix: npt.NDArray[np.bool_],
    ) -> list[list[int]]:
        """
        Find connected components in the graph represented by the adjacency matrix.
        In other words, groups the lines that belong to the same table on the page.

        Args:
            adjacency_matrix: the matrix representing which line intersects which line.
                (output of build_adjacency_matrix()).

        Returns:
            list[list[int]]: a list of list of indexes representing grouped lines
        """
        n = len(adjacency_matrix)
        visited = np.zeros(n, dtype=bool)
        components: list[list[int]] = []

        for start in range(n):
            if not visited[start]:
                stack = [start]
                component: list[int] = []
                while stack:
                    node = stack.pop()
                    if not visited[node]:
                        visited[node] = True
                        component.append(node)
                        neighbors = np.where(adjacency_matrix[node])[0]
                        stack.extend(neighbors)
                components.append(component)

        return components

    def group_lines_by_table(
        self, coordinates: npt.NDArray[np.float32]
    ) -> list[npt.NDArray[np.float32]]:
        """Considering a list of matrix coordinates of shape (n_lines, 4),
        with each row reprensenting (x1, y1, x2, y2) coordinates of the lines,
        groups the lines together if they belong to the same table.

        Args:
            coordinates (npt.NDArray[np.float32]): the coordinates of lines.

        Returns:
            list[npt.NDArray[np.float32]]: the groups of lines coordinates, grouped by table.
        """
        adjacency_matrix = self._build_adjacency_matrix(coordinates)
        indexes_of_groups = TableFinder._connected_components(adjacency_matrix)
        return [coordinates[group] for group in indexes_of_groups if group]

    @staticmethod
    def _get_cells(
        intersections: npt.NDArray[np.float32],
        lines_coordinates: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """Considering a list of intersection points in a table,
        and a list of detected lines, gets all the cells of table.

        Args:
            intersections (npt.NDArray[np.float32]): a list of intersection points,
                as an array of shape (n_points, 2) of (x,y) coordinates.
            lines_coordinates (npt.NDArray[np.float32]): an array of shape (n_lines, 4)
                where each row contains coordinates of the x0, y0, x1, y1 of each line.

        Returns:
            npt.NDArray[np.float32]: the list of cells, as an array of shape
                (n_cells, 4) where each row is (x0, y0, x1, y1) coordinates.
        """
        cells_coords = TableFinder.get_table_cells_without_merged_cells(intersections)
        valid_cell_borders_mask = TableFinder._get_cell_borders_validity(
            cells_coords, lines_coordinates
        )
        if valid_cell_borders_mask.all():  # No merged cells
            return cells_coords
        # Remove invalid cells
        valid_cells_mask = valid_cell_borders_mask.all(axis=1)
        valid_cells = cells_coords[valid_cells_mask]
        # For invalid cells, find the merged cells
        topleft_corners_of_merged_cells = cells_coords[~valid_cells_mask][:, :2]
        line_combinations_coords = TableFinder._get_recombined_lines(lines_coordinates)
        merged_cells = [
            TableFinder._get_merged_cell(point, intersections, line_combinations_coords)
            for point in topleft_corners_of_merged_cells
        ]
        merged_cells = np.unique(np.vstack(merged_cells), axis=0)

        return np.unique(np.vstack((merged_cells, valid_cells)), axis=0)

    @staticmethod
    def get_table_cells_without_merged_cells(
        intersections: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """From line intersections, build cells as if
        no cells where merged.

        Args:
            intersections (npt.NDArray[np.float32]): a list of intersection points,
                as an array of shape (n_points, 2) of (x,y) coordinates.

        Returns:
            npt.NDArray[np.float32]: the list of cells, as an array of shape
                (n_cells, 4) where each row is (x0, y0, x1, y1) coordinates.
        """
        # Get unique x and y positions from intersections
        xs, ys = np.unique(intersections[:, 0]), np.unique(intersections[:, 1])
        # Generate cells
        x1, x2 = np.repeat(xs[:-1], len(ys) - 1), np.repeat(xs[1:], len(ys) - 1)
        y1, y2 = np.tile(ys[:-1], len(xs) - 1), np.tile(ys[1:], len(xs) - 1)
        cells = np.column_stack([x1, y1, x2, y2])

        return cells

    @staticmethod
    def _get_merged_cell(
        topleft_point: npt.NDArray[np.float32],
        intersections: npt.NDArray[np.float32],
        lines_combinations: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """Considering a point coordinate that would be topleft corner of a cell,
        and a list of intersections points, and a list of lines,
        get its cells coordinates corresponding to the point.

        Args:
            topleft_point (npt.NDArray[np.float32]): an array of size (1, 2) reprensenting
                the (x, y) coordinates of what is supposedly the topleft corner of a cell.
            intersections (npt.NDArray[np.float32]): a list of intersection points,
                as an array of shape (n_points, 2) of (x,y) coordinates.
            lines_combinations: npt.NDArray[np.float32]: an array representing
                all the combinations of lines that can be made for the detected lines.

        Returns:
            npt.NDArray[np.float32] : An array of shape (0, 4) or (1, 4) representing the cell coordinates
                corresponding to provided points.
        """
        cells = TableFinder._get_potential_cells_for_point(topleft_point, intersections)
        if cells.size:
            cells = TableFinder._remove_cells_with_invalid_borders(
                cells, lines_combinations
            )
            cells = TableFinder._remove_nesting_cells(cells)
            assert cells.shape[0] in [
                0,
                1,
            ], "Error, only 1 or 0 cells should be returned."
        else:
            cells = np.empty((0, 4), dtype=np.float32)

        return cells

    @staticmethod
    def _get_potential_cells_for_point(
        point: npt.NDArray[np.float32], intersections: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Considering a point and a list of intersection points,
        gets all the possible cells that can be made for that point.

        Example:
        o11---o12---o13
         |    |     |
        o21---o22---o23
         |    |     |
        o31---o32---o33

        Considering the above table, for point o11,
        the potention cells would be delimited by
        [o11, o22], [o11, o23], [o11, o32], [o11, o33]

        Args:
            point (npt.NDArray[np.float32]): point coordinates (array of shape (1, 2))
            intersections (npt.NDArray[np.float32]): a list of intersection points,
                as an array of shape (n_points, 2) of (x,y) coordinates.

        Returns:
            npt.NDArray[np.float32]: the list of potential cell, as an array of shape
                (n_cells, 4) where each row is (x0, y0, x1, y1) coordinates.
        """
        potential_points = intersections[
            (intersections[:, 0] > point[0]) & (intersections[:, 1] > point[1])
        ]
        return np.concatenate(
            [np.broadcast_to(point, (potential_points.shape[0], 2)), potential_points],
            axis=1,
        )

    @staticmethod
    def _remove_cells_with_invalid_borders(
        cells_coords: npt.NDArray[np.float32],
        lines_coordinates: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """Considering a list of cells, and a list of detected lines,
        remove the cells for which the borders are not along a line.
        In other words, verifies that all borders of each cell are actually along a detected line.

        Args:
            cells_coords (npt.NDArray[np.float32]): an array of shape (n_cells, 4) where each row contains
                the (x1, y1, x2, y2) coordinates of the cells.
            lines_coordinates (npt.NDArray[np.float32]): an array of shape (n_lines, 4) where each row contains
                coordinates of the x0, y0, x1, y1 of each line.
        """
        valid_cell_borders_mask = TableFinder._get_cell_borders_validity(
            cells_coords, lines_coordinates
        )
        valid_cells = cells_coords[valid_cell_borders_mask.all(axis=1)]

        return valid_cells

    @staticmethod
    def _get_cell_borders_validity(
        cells_coords: npt.NDArray[np.float32],
        lines_coordinates: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.bool_]:
        """Considering a list of cells, and a list of detected lines,
        returns an array of shape (n_cells, 4) where each row represents
        whether or not each border of cells exists in lines_coordinates.
        Each row represents the validity of left, top, right, and bottom border in that order.

        Args:
            cells_coords (npt.NDArray[np.float32]): an array of shape (n_cells, 4) where each row contains
                the (x1, y1, x2, y2) coordinates of the cells.
            lines_coordinates (npt.NDArray[np.float32]): an array of shape (n_lines, 4) where each row contains
                coordinates of the x0, y0, x1, y1 of each line.
        """
        x0, y0, x1, y1 = TableFinder._bounding_boxes(cells_coords)
        # build array of shape (n_cells, 4, 4), in which we have the coords of the 4 borders of each cell.
        left_lines = np.array([x0, y0, x0, y1]).T
        top_lines = np.array([x0, y0, x1, y0]).T
        right_lines = np.array([x1, y0, x1, y1]).T
        bottom_lines = np.array([x0, y1, x1, y1]).T

        cells_borders = np.stack(
            [left_lines, top_lines, right_lines, bottom_lines], axis=1
        )
        lines_coordinates_set = set(map(tuple, lines_coordinates))
        valid_borders_mask = np.array(
            [
                [tuple(border) in lines_coordinates_set for border in cell_borders]
                for cell_borders in cells_borders
            ],
            np.bool_,
        )

        return valid_borders_mask

    @staticmethod
    def _remove_nesting_cells(
        cells_coords: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """Considering a list of cells, one keep the cells that do not contain
        any other cells.

        Args:
            cells_coords (npt.NDArray[np.float32]): an array of shape (n_cells, 4) where each row contains
                the (x1, y1, x2, y2) coordinates of the cells.

        Returns:
            npt.NDArray[np.float32]: cells coordinates of cells that do not contain any other cell.
        """
        x0, y0, x1, y1 = TableFinder._bounding_boxes(cells_coords)
        # Compare each coord of each box with other boxes
        x0_mask = x0[:, np.newaxis] <= x0
        y0_mask = y0[:, np.newaxis] <= y0
        x1_mask = x1[:, np.newaxis] >= x1
        y1_mask = y1[:, np.newaxis] >= y1
        containment_mask = x0_mask & y0_mask & x1_mask & y1_mask
        # Ignore self-containment by setting diagonal to False
        np.fill_diagonal(containment_mask, False)
        # Find cells that do NOT contain any other cell
        containment_mask = ~np.any(containment_mask, axis=1)

        return cells_coords[containment_mask]

    @staticmethod
    def point_is_on_line(
        point: tuple[float, float],
        line: tuple[float, float, float, float],
        tolerance: float = 1e-6,
    ) -> bool:
        """Check if a point is on a line segment."""
        x1, y1, x2, y2 = line
        x, y = point
        # Check if the point lies on the line using vector cross product
        if abs((y - y1) * (x2 - x1) - (x - x1) * (y2 - y1)) > tolerance:
            return False
        # Check if the point is within the line segment bounds
        if min(x1, x2) <= x <= max(x1, x2) and min(y1, y2) <= y <= max(y1, y2):
            return True

        return False

    @staticmethod
    def subdivide_lines(
        lines: npt.NDArray[np.float32], intersections: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Subdivides all the lines that cross an intersection point.

        Exemple : o1---o2---o3---o4. Considering we have line (o1-o4),
        this method could return lines [(o1-o2),(o2-o3),(o3-o4)]

        Args:
            lines_coordinates (npt.NDArray[np.float32]): an array of shape (n_lines, 4) where each row contains
                coordinates of the x0, y0, x1, y1 of each line.
            intersections (npt.NDArray[np.float32]): a list of intersection points,
                as an array of shape (n_points, 2) of (x,y) coordinates.

        Returns:
            npt.NDArray[np.float32]: the cooridiantes of subdivided lines.
        """
        new_lines: list[tuple[float, float, float, float]] = []

        for line in lines:
            split_points: list[tuple[float, float]] = []

            # Check each intersection to see if it lies on the line
            for intersection in intersections:
                if TableFinder.point_is_on_line(intersection, line):
                    split_points.append(intersection)
            # Sort the split points in the order they appear on the line
            x0, y0 = line[:2]
            split_points.sort(
                key=lambda p, x0=x0, y0=y0: abs(p[0] - x0) + abs(p[1] - y0)
            )
            # Subdivide the line at each of these split points
            x1, y1 = line[:2]
            for x, y in split_points:
                new_lines.append([x1, y1, x, y])
                x1, y1 = x, y
            new_lines.append([x1, y1, line[2], line[3]])
        # Get uniues lines and remove lines that are actually points
        new_lines = np.unique(np.array(new_lines), axis=0)
        return new_lines[
            (new_lines[:, 0] != new_lines[:, 2]) | (new_lines[:, 1] != new_lines[:, 3])
        ]

    @staticmethod
    def _get_recombined_lines(
        lines_coordinates: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """Gets the coordinates of all te lines that can be built
        based on recombination of continuous segments along x and y axis.

        Exemple:
        o11---o12---o13---o14
        Considering the above segments, this method would return
        all segments provided as arg + the coordinates of [o1, o3], [o1, o4], [o2, o4]
        Note: works for both horizontal and vertical axes.

        Args:
            lines_coordinates (npt.NDArray[np.float32]): A numpy array of horizontal and vertical segments
                of shape (n_lines, 4) where each row is [x0, y0, x1, y1] coordinates of a line.

        Returns:
            npt.NDArray[np.float32]: An array of shape (n, 4) containing
            all the coordinates of the segments + every possible recombination.
        """
        combined_lines = lines_coordinates
        n_prev_combined_lines = -1

        while len(combined_lines) != n_prev_combined_lines:
            n_prev_combined_lines = len(combined_lines)
            new_combinations: list[list[float]] = []
            # Build dictionaries for quick lookup of start and end points
            start_to_indices: dict[tuple[float, float], list[int]] = {}
            for idx, line in enumerate(combined_lines):
                start_point = (line[0], line[1])
                start_to_indices.setdefault(start_point, []).append(idx)

            for idx, (x0, y0, x1, y1) in enumerate(combined_lines):
                # Look for lines that start where the current line ends
                end_point = (x1, y1)
                if end_point in start_to_indices:
                    # Combine with all lines starting at the end point
                    for next_idx in start_to_indices[end_point]:
                        # only pay attention to vertical/horizontal lines
                        if (
                            combined_lines[next_idx][2] == x0
                            or combined_lines[next_idx][3] == y1
                        ):
                            new_line = [
                                x0,
                                y0,
                                combined_lines[next_idx][2],
                                combined_lines[next_idx][3],
                            ]
                            new_combinations.append(new_line)

            if new_combinations:
                # Add new combinations and remove duplicates
                combined_lines = np.unique(
                    np.vstack(
                        (combined_lines, np.array(new_combinations, dtype=np.float32))
                    ),
                    axis=0,
                )

        return combined_lines

    @staticmethod  # TODO : investigate if possible to replace _get_recombined_lines with this one
    def _get_recombined_lines_v2(
        lines_coordinates: npt.NDArray[np.float32],
    ) -> npt.NDArray[np.float32]:
        """WARNING : IMPLEMENTATION DOESN'T WORK
        I leave it here for further investigation as it might be faster and easier to read
        than current used implementation : _get_recombined_lines.
        """
        prev_n_lines = -1  # dummy to go in while loop
        while prev_n_lines < lines_coordinates.shape[0]:
            prev_n_lines = lines_coordinates.shape[0]
            start_points = lines_coordinates[:, :2]
            end_points = lines_coordinates[:, 2:]
            # Get mask where start point and endpoints match
            mask = (start_points[:, np.newaxis, :] == end_points[np.newaxis, :, :]).all(
                axis=2
            )  # PROBLEM WITH THIS LINE
            new_lines = np.hstack(
                [start_points[mask.any(axis=0)], end_points[mask.any(axis=1)]]
            )
            new_lines = new_lines[
                (new_lines[:, 0] == new_lines[:, 2])
                | (new_lines[:, 1] == new_lines[:, 3])
            ]  # only vertical and horizontal lines
            lines_coordinates = np.unique(
                np.vstack([new_lines, lines_coordinates]), axis=0
            )

        return lines_coordinates

    @staticmethod
    def _table_sanity_check(cells_coords: npt.NDArray[np.float32]) -> bool:
        """Perfoms a sanity check of the parsed table.
        Checks that the area of the table is equal to the sum of the area of its cells.

        Args:
            cells (npt.NDArray[np.float32]): the cells of the table detected.
                Array of shape (n_cells, 4) where each row is x0, y0, x1, y1 coordinates of the cell.

        Returns:
            bool: returns True if the sanity check is passed, and False if the table
                is likely to be a false detection.
        """
        x0, y0, x1, y1 = (
            cells_coords[:, 0],
            cells_coords[:, 1],
            cells_coords[:, 2],
            cells_coords[:, 3],
        )

        table_area = (np.max(x1) - np.min(x0)) * (np.max(y1) - np.min(y0))
        cells_areas = np.abs((x1 - x0) * (y1 - y0))
        total_cells_area = np.sum(cells_areas)

        return np.isclose(table_area, total_cells_area, rtol=0.001)



from pathlib import Path
import html

import mammoth  # type: ignore : No stub files

from ...core.components import MarkdownDoc
from ..html.html_parser import HTMLParser


class DocxParser(HTMLParser):

    def parse_string(self, string: str) -> MarkdownDoc:
        """Parses a HTML-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            string (str): the markdown formatted string

        Returns:
            MarkdownDoc: the parsed document. Can be fed to chunker.
        """
        formatted_string = DocxParser.apply_markdownify(string)
        formatted_string = DocxParser.cleanup_string(formatted_string)

        return MarkdownDoc.from_string(formatted_string)

    def parse_file(self, filepath: str) -> MarkdownDoc:
        """Reads and parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            filepath (FilePath): the path to a .html file

        Returns:
            MarkdownDoc: the parsed document. Can be fed to chunker.
        """
        html_string = DocxParser.read_file(filepath)

        return self.parse_string(html_string)

    @staticmethod
    def read_file(filepath: str) -> str:
        """Reads a Markdown file

        Args:
            filepath (str): the path to the HTML file.

        Returns:
            str: the HTML string.
        """
        path = Path(filepath)
        if path.suffix != ".docx":
            raise ValueError("Only .docx files can be passed to DocxParser.")
        with path.open("rb") as file:
            html_string = mammoth.convert_to_html(file).value  # type: ignore : Type of input is unknown
            html_string = html.unescape(html_string)

        return html_string



import json
from pathlib import Path
from typing import Any

import nbformat
from pydantic import ValidationError

from ...core.components import MarkdownDoc
from ..abstract_parser import AbstractParser
from .schemas import JupyterNotebookContent


class JupyterNotebookParser(AbstractParser):
    """Class used to parse jupyter notebooks (.ipynb)."""

    def __init__(self, include_code_cells_outputs: bool = False) -> None:
        """Initializes the class.

        Args:
            include_code_cells_outputs (bool) : Whether or not the cells' output
                should be parsed as well. If False, they will be ignored. Default to False.
        """
        self.include_code_cells_outputs = include_code_cells_outputs

    def parse_file(self, filepath: str) -> MarkdownDoc:
        """Chunks a notebook .ipynb file.

        Args:
            filepath (str): the path to the file to parse.

        Returns:
            MarkDownDoc : the parsed markdown object.
        """
        file_content = self.read_file(filepath)
        md_string = self._parse_notebook_content(file_content)

        return MarkdownDoc.from_string(md_string)

    def parse_string(self, string: str) -> MarkdownDoc:
        """Parses the string considering it is a notebook content.

        Args:
            string (str): the string representing the notebook content. Assumed to be a dumped
                version of the JSON formatted notebook.

        Returns:
            MarkdownDoc: the formatted markdown document.
        """
        try:
            content = nbformat.reads(string, 4)  # type: ignore
            content = JupyterNotebookContent(**content)
        except ValidationError as e:
            raise e

        md_string = self._parse_notebook_content(content)

        return MarkdownDoc.from_string(md_string)

    @staticmethod
    def read_file(filepath: str) -> JupyterNotebookContent:
        """Reads a .ipynb file and returns its
        content as a json dict.

        Args:
            filepath (str): path to the file

        Returns:
            dict[str, Any]: the json content of the ipynb file
        """
        path = Path(filepath)
        if path.suffix != ".ipynb":
            raise ValueError(
                "Only .ipynb files can be passed to JupyterNotebookParser."
            )
        try:
            content = nbformat.read(filepath, 4)  # type: ignore
            content = JupyterNotebookContent(**content)
        except ValidationError as e:
            raise e

        return content

    def _parse_notebook_content(self, notebook_content: JupyterNotebookContent) -> str:
        """Parses the content of the notebook.

        Args:
            notebook_content (JupyterNotebookContent): the content of the notebook.

        Returns:
            str: the markdown string parsed from the notebook content
        """
        kernel_language = (
            notebook_content.metadata["language_info"]["name"]
            if notebook_content.metadata
            else ""
        )
        md_string = ""
        for cell in notebook_content.cells:
            match cell.cell_type:
                case "markdown" | "raw":
                    md_string += cell.source + "\n\n"
                case "code":
                    language = (
                        cell.metadata.get("vscode", {}).get("languageId", None)
                        or kernel_language
                    )
                    md_string += (
                        "```" + language + "\n" + "".join(cell.source) + "\n```\n\n"
                    )
                    if self.include_code_cells_outputs and cell.outputs:
                        md_string += (
                            "\n\n".join(
                                JupyterNotebookParser._parse_cell_output(output)
                                for output in cell.outputs
                            )
                            + "\n\n"
                        )

        return md_string

    @staticmethod
    def _parse_cell_output(cell_output: dict[str, Any]) -> str:
        """Parses a cell output to get a string from it.
        The format of a notebook's cell output's is described here : https://nbformat.readthedocs.io/en/latest/format_description.html#code-cell-outputs

        Args:
            cell_output (dict[str, Any]): a dictionnary representing a cell's output.

        Returns:
            str : the cell output as a string.
        """

        match cell_output["output_type"]:
            case "execute_result" | "display_data":
                text_output = ""
                for key, value in cell_output.get("data", {}).items():
                    if key == "application/json" or key.endswith("+json"):
                        text_output += json.dumps(value) + "\n\n"
                    elif key == "text/plain":
                        text_output += value + "\n\n"
                    elif key.startswith("image/"):
                        pass  # Do not include images as they are Base64 encoded ?
                return text_output.strip()
            case "stream":
                return cell_output["text"].strip()
            case _:
                pass



import csv
import re
from io import StringIO
from typing import Literal

import pandas as pd

from ...core.components import MarkdownDoc
from ..abstract_parser import AbstractParser


class CSVParser(AbstractParser):
    """Parser for Comma-Separated Values file (.csv)"""

    def __init__(
        self,
        csv_delimiter: str | None = None,
        output_format: Literal["markdown_table", "json_lines"] = "json_lines",
    ) -> None:
        """Initializes a sheet parser

        Args:
            csv_delimiter (str | None, optional): The delimiter to consider to parse the .csv files.
                If None, we will try to guess what the delimiter is. Defaults to None.
            output_format (Literal[&quot;markdown_table&quot;, &quot;json_lines&quot;], optional): the output format of the parsed document.
                - markdown_table : uses tabula to build a markdown-formatted table.
                - json_lines : each row of the table will be output as a JSON line. NOTE : consumes way more tokens as column names are repeated at each row. But easier to read for LLMs.
                Defaults to "json_lines".
        """
        self.csv_delimiter = csv_delimiter
        self.output_format = output_format

    def parse_file(self, filepath: str) -> MarkdownDoc:
        """Parses a csv file to markdown.

        Args:
            filepath (str): the path to the csv file.

        Returns:
            MarkdownDoc: the markdown-formatted csv.
        """
        csv_string = self.read_file(filepath)

        return self.parse_string(csv_string)

    def parse_string(self, string: str) -> MarkdownDoc:
        """Parses a string representing a csv file to markdown.

        Args:
            string (str): the csv-formatted string.

        Returns:
            MarkdownDoc: the markdown-formatted csv.
        """
        delimiter = self.csv_delimiter or CSVParser._detect_csv_delimiter(string)
        df = pd.read_csv(StringIO(string), delimiter=delimiter)  # type: ignore | missing typing in pandas
        match self.output_format:
            case "markdown_table":
                md_string = CSVParser.convert_df_to_markdown_table(df)
            case "json_lines":
                md_string = df.to_json(orient="records", force_ascii=False, lines=True)  # type: ignore | df.to_json() -> str
            case _:
                raise ValueError(
                    f"Invalid value for argument 'output_format' : expected one of ['markdown_table', 'json_lines]. Got '{self.output_format}'."
                )

        return MarkdownDoc.from_string(md_string)

    def read_file(self, filepath: str) -> str:
        """Read the provided filepath. For a list of handled filetypes,
        refer to https://pandas.pydata.org/docs/reference/api/pandas.read_excel.html.

        Args:
            filepath (str): path to the file.

        Returns:
            str: the csv file content as a string.
        """
        if not filepath.lower().endswith(".csv"):
            raise ValueError("Only .csv files can be passed to CSVParser.")
        with open(filepath, "r", encoding="utf8") as file:
            csv_string = "".join(file.readlines())

        return csv_string

    @staticmethod
    def _detect_csv_delimiter(csv_string: str, n_sample_lines: int = 5) -> str:
        """Detect the delimiter used in a CSV file.

        Args:
            csv_string (str): the csv file as a string.
            n_sample_lines (int): the amount of lines to consider for guessing the separator.
                Higher number may increase inference time.

        Returns:
            str : the delimiter
        """
        sample = "\n".join(csv_string.split("\n")[:n_sample_lines])
        sniffer = csv.Sniffer()
        try:
            dialect = sniffer.sniff(sample)
            return dialect.delimiter
        except csv.Error as e:
            raise ValueError(
                "Could not detect the delimiter. You may want to set delimiter manually using SheetParser(csv_delimiter='<mydelimiter>') before parsing the file."
            ) from e

    @staticmethod
    def convert_df_to_markdown_table(df: pd.DataFrame) -> str:
        """Converts a DataFrame to markdown.
        Wraps tabula's method pd.DataFrame.to_markdown()
        between pre and post processing.
        Preprocess :
        - Remove \n in text columns
        PostProcess :
        - Replace multiple spaces with 2 spaces.

        Args:
            df (pd.DataFrame): the dataframe to convert.

        Returns:
            str: a markdown formatted table.
        """
        dtypes = df.apply(  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
            lambda x: pd.api.types.infer_dtype(x, skipna=True)  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
        )
        string_cols = dtypes[dtypes == "string"].index  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
        df[string_cols] = df[string_cols].apply(lambda x: x.str.replace("\n", " "))  # type: ignore | x: pd.Series[Any] -> pd.Series[str]
        md_string = df.to_markdown(index=False)
        md_string = re.sub(r"\s{3,}", "  ", md_string)
        md_string = re.sub(r"-{3,}", "---", md_string)

        return md_string

    @staticmethod
    def convert_df_to_json_lines(df: pd.DataFrame) -> str:
        """Converts a DataFrame to json lines.

        Args:
            df (pd.DataFrame): the dataframe to convert.

        Returns:
            str: the json lines.
        """
        return df.to_json(orient="records", force_ascii=False, lines=True)  # type: ignore | df.to_json() -> str



from abc import ABC, abstractmethod
from typing import Any

from pydantic import validate_call


class AbstractParser(ABC):
    """A parser is meant to parse either a string
    of a document provided as a file path.
    Either way, its output must be ingestable by a chunker.
    IMPORTANT : The types specified in the 2 abstract methods
    must be the same.
    """

    @validate_call
    @abstractmethod
    def parse_string(self, string: str) -> Any:
        pass

    @validate_call
    @abstractmethod
    def parse_file(self, filepath: str) -> Any:
        pass



from typing import Any, Literal

from ..core.components import Chunk, MarkdownDoc, MarkdownLine, TocTree
from ..decorators.decorators import timeit, validate_args
from .abstract_chunker import AbstractChunker


class MarkdownChunker(AbstractChunker):
    max_headers_to_use: str
    max_chunk_word_count: int
    hard_max_chunk_word_count: int
    min_chunk_word_count: int

    def __init__(
        self,
        max_headers_to_use: Literal["h1", "h2", "h3", "h4", "h5", "h6"] = "h4",
        max_chunk_word_count: int = 200,
        hard_max_chunk_word_count: int = 400,
        min_chunk_word_count: int = 15,
        hard_max_chunk_token_count: int | None = None,
        tokenizer: Any | None = None,
    ) -> None:
        """Initialize a Markdown chunker

        Args:
            max_headers_to_use (MaxHeadersToUse) : The maximum header level to consider (included).
                Headers with level lower than this wont be used to split chunks.
                For example, if 'h4' is set, then 'h5' and 'h6' headers won't be used.
                Must be a string of type 'hx' with x being the title level ranging from 1 to 6.
            max_chunk_word_count (int) : The maximum size a chunk can be (in words).
                It is a SOFT limit, meaning that chunks bigger that this will be chunked only if lower level headers if any are available."
            hard_max_chunk_word_count (int) : The true maximum size a chunk can be (in word).
                It is a HARD limit, meaning that chunks bigger by this limit will be split into subchunks.
            min_chunk_word_count (int) : The minimum size a chunk can be (in words).
                Chunks smaller than this will be discarded.
            hard_max_chunk_token_count (None | int) : The true maximum size a chunk can be (in tokens). If None, no token-based splitting will be done.
                It is a HARD limit, meaning that chunks bigger by this limit will be split into subchunks that are equivalent in terms of tokens count.
            tokenizer (Any | None) : The tokenizer to use. Can be any instance of a class that has 'encode' method such as tiktoken.
        """
        self.max_headers_to_use = max_headers_to_use
        self.max_chunk_word_count = max_chunk_word_count
        self.hard_max_chunk_word_count = hard_max_chunk_word_count
        self.min_chunk_word_count = min_chunk_word_count
        self.hard_max_chunk_token_count = hard_max_chunk_token_count
        self.tokenizer = tokenizer

    @timeit
    @validate_args
    def chunk(self, content: MarkdownDoc) -> list[Chunk]:
        """Chunks a parsed Markdown document.

        Args:
            content (MarkdownDoc): the markdown document to chunk.
                Might be the output of a chunknorris.Parser.

        Returns:
            list[Chunk]: the chunks.
        """
        toc_tree = self.get_toc_tree(content.content)
        chunks = self.get_chunks(toc_tree)

        return chunks

    def get_toc_tree(
        self,
        md_lines: list[MarkdownLine],
    ) -> TocTree:
        """Builds the table of content tree based on header.

        Args:
            md_lines (list[MarkdownLines]): the markdown lines.

        Returns:
            TocTree: the table of content.
        """
        max_header_level_to_use = int(self.max_headers_to_use[1])
        dummy_line = MarkdownLine("", line_idx=-1)
        tree = TocTree(title=dummy_line)
        current_node = tree
        id_cntr = 0
        for line in md_lines:
            if line.is_header and line.get_header_level() <= max_header_level_to_use:
                while (
                    current_node.parent is not None
                    and line.get_header_level() <= current_node.title.get_header_level()
                ):
                    current_node = current_node.parent
                current_node.add_child(
                    TocTree(
                        title=line,
                        id=id_cntr,
                    )
                )
                current_node = current_node.children[-1]
                id_cntr += 1
            else:
                current_node.content.append(line)

        return tree

    def get_chunks(self, toc_tree: TocTree) -> list[Chunk]:
        """Wrapper that build the chunk's texts, check
        that they fit in size, replace links formatting.

        Args:
            toc_tree (TocTree): the toc tree of the document.

        Returns:
            Chunks: the chunks text, formatted.
        """
        chunks = self.build_chunks(toc_tree)
        chunks = self.split_big_chunks_wordbased(chunks)
        chunks = self.split_big_chunks_tokenbased(chunks)
        chunks = self.remove_small_chunks(chunks)

        return chunks

    def build_chunks(
        self,
        toc_tree_element: TocTree,
        already_ok_chunks: list[Chunk] | None = None,
    ) -> list[Chunk]:
        """Uses the toc tree to build the chunks. Uses recursion.
        Method :
        - build the chunk (= titles from sections above + section content + content of subsections)
        - if the chunk is too big:
            - save the section as title + content (if section has content)
            - subdivide section recursively using subsections
        - else save it as is

        Args:
            toc_tree_element (TocTree): the TocTree for which the chunk should be build
            already_ok_chunks (Chunks, optional): the chunks already built.
                Used for recursion. Defaults to None.

        Returns:
            Chunks: list of chunk's texts.
        """
        if already_ok_chunks is None:
            already_ok_chunks = []

        current_chunk = MarkdownChunker._build_chunk(toc_tree_element)

        if current_chunk.word_count > self.max_chunk_word_count:
            if any(line.text.strip() for line in toc_tree_element.content):
                parent_headers = MarkdownChunker.get_parents_headers(toc_tree_element)
                current_chunk = Chunk(
                    headers=parent_headers + [toc_tree_element.title],
                    content=toc_tree_element.content,
                    start_line=toc_tree_element.title.line_idx,
                )
                already_ok_chunks.append(current_chunk)
            for child in toc_tree_element.children:
                already_ok_chunks = self.build_chunks(child, already_ok_chunks)
        else:
            already_ok_chunks.append(current_chunk)

        return already_ok_chunks

    @staticmethod
    def _build_chunk(toc_tree_element: TocTree) -> Chunk:
        """Builds a chunk by apposing the text of headers
        and recursively getting the content of children.

        Args:
            toc_tree_element (TocTree): the toc tree element.

        Returns:
            str: the chunk content. parent's headers + content.
        """
        parent_headers = MarkdownChunker.get_parents_headers(toc_tree_element)
        content = MarkdownChunker._build_chunk_content(toc_tree_element)

        return Chunk(
            headers=parent_headers,
            content=content,
            start_line=toc_tree_element.title.line_idx,
        )

    @staticmethod
    def _build_chunk_content(toc_tree_element: TocTree) -> list[MarkdownLine]:
        """Builds a chunk content (i.e without headers above)
        from a toc tree. It uses the toc tree's content, and recursively
        adds the header and content of its children.

        Args:
            toc_tree_element (TocTree): the toc tree (or element of toc tree).

        Returns:
            list[MarkdownLine]: the list of lines that belong to the chunk content (without the headers of parents).
        """
        content = [toc_tree_element.title] + toc_tree_element.content
        for child in toc_tree_element.children:
            content.extend(MarkdownChunker._build_chunk_content(child))

        return content

    @staticmethod
    def get_parents_headers(toc_tree_element: TocTree) -> list[MarkdownLine]:
        """Gets a list of the titles that are parent
        of the provided toc tree element. The list
        is ordered in descending order in terms of header level.

        Args:
            toc_tree_element (TocTree): the toc tree element.

        Returns:
            list[MarkdownLine]: the list of line that represent the parent's headers.
        """
        headers: list[MarkdownLine] = []
        while toc_tree_element.parent:
            toc_tree_element = toc_tree_element.parent
            headers.append(toc_tree_element.title)
        # remove empty string header, such as root header
        headers = [h for h in headers if h.text]

        return list(reversed(headers))

    def remove_small_chunks(self, chunks: list[Chunk]) -> list[Chunk]:
        """Removes chunks that have less words than the specified limit.

        Args:
            chunks (Chunks): the list of chunks.

        Returns:
            Chunks: the chunks with more words than the specified threshold.
        """
        return [c for c in chunks if c.word_count >= self.min_chunk_word_count]

    def split_big_chunks_wordbased(
        self,
        chunks: list[Chunk],
    ) -> list[Chunk]:
        """Splits the chunks that are too big.
        You may consider passing the kwarg "hard_max_chunk_word_count"
        to specify the limit size of the chunk (in words).

        Args:
            chunks (Chunks): The chunks obtained from the get_chunks() method.

        Returns:
            Chunks: the chunks, with big chunks splitted into smaller chunks.
        """
        splitted_chunks: list[Chunk] = []
        for chunk in chunks:
            if chunk.word_count < self.hard_max_chunk_word_count:
                splitted_chunks.append(chunk)
            else:
                splitted_chunk = self._split_with_newlines(chunk)
                splitted_chunks.extend(splitted_chunk)

        return splitted_chunks

    def _split_with_newlines(
        self,
        chunk: Chunk,
    ) -> list[Chunk]:
        """Split chunks based on newlines. Adds the
        chunk titles at the beginning of each chunk.

        Args:
            chunk (Chunk): the chunk to split.
        """
        split_count = (chunk.word_count // self.hard_max_chunk_word_count) + 1
        split_word_size = chunk.word_count // split_count

        chunks: list[Chunk] = []
        line_buffer: list[MarkdownLine] = []
        headers_buffer = [None] * 7
        prev_buffer = headers_buffer.copy()
        for line in chunk.content:
            line_buffer.append(line)
            if line.is_header:
                header_level = line.get_header_level()
                headers_buffer[header_level:] = [None] * len(
                    headers_buffer[header_level:]
                )
                headers_buffer[header_level] = line
                continue
            # Do not split the chunk at this line to preserve tables, bullet points list and code blocks.
            elif line.is_bullet_point or line.isin_code_block or line.isin_table:
                continue
            current_word_count = len("\n".join(lb.text for lb in line_buffer).split())
            if current_word_count > split_word_size:
                chunks.append(
                    MarkdownChunker._create_new_chunk_from_lines(
                        chunk.headers + list(filter(None, prev_buffer)), line_buffer
                    )
                )
                line_buffer = []
                prev_buffer = headers_buffer
        if line_buffer:
            chunks.append(
                MarkdownChunker._create_new_chunk_from_lines(chunk.headers, line_buffer)
            )

        return chunks

    def split_big_chunks_tokenbased(
        self,
        chunks: list[Chunk],
    ) -> list[Chunk]:
        """Splits the chunks that are too big considering the provided tokenizer.

        Args:
            chunks (list[Chunk]): the chunks to split.

        Raises:
            ValueError: if the tokenizer is not provided.
            ValueError: if the tokenizer does not have 'encode' method.

        Returns:
            list[Chunk]: the chunks, with big chunks splitting into smaller chunks.
        """
        if self.hard_max_chunk_token_count is None:
            return chunks
        if self.tokenizer is None:
            raise ValueError(
                "Tokenizer is required when hard_max_chunk_token_count is set"
            )
        if not hasattr(self.tokenizer, "encode"):
            raise ValueError(
                "Tokenizer must have 'encode' methods that takes a string a intput and return a list of tokens."
            )

        splitted_chunks: list[Chunk] = []
        for chunk in chunks:
            splitted_chunks.extend(
                self._split_with_tokenizer(
                    chunk, self.tokenizer, self.hard_max_chunk_token_count
                )
            )
        return splitted_chunks

    @staticmethod
    def _split_with_tokenizer(
        chunk: Chunk, tokenizer: Any, hard_max_chunk_token_count: int
    ) -> list[Chunk]:
        """Split a chunk using the provided tokenizer and the hard_max_chunk_token_count."""
        chunk_tokens_count = len(tokenizer.encode(chunk.get_text()))
        if chunk_tokens_count < hard_max_chunk_token_count:
            return [chunk]
        else:
            n_splits = chunk_tokens_count // hard_max_chunk_token_count + 1
            tokens_per_split = chunk_tokens_count // n_splits
            line_buffer: list[MarkdownLine] = []
            initial_token_count = len(
                tokenizer.encode("\n\n".join([header.text for header in chunk.headers]))
            )
            current_token_count = initial_token_count
            new_chunks: list[Chunk] = []
            for line in chunk.content:
                line_token_count = len(tokenizer.encode(line.text + "\n"))
                # if line is too big for a single chunk -> split line and make chunks
                if line_token_count > hard_max_chunk_token_count:
                    if line_buffer:
                        new_chunks.append(
                            MarkdownChunker._create_new_chunk_from_lines(
                                chunk.headers, line_buffer
                            )
                        )
                        line_buffer = []
                    # split line in multiple chunks
                    new_chunks.extend(
                        MarkdownChunker._split_line_into_chunks(
                            line,
                            line_token_count,
                            hard_max_chunk_token_count,
                            chunk.headers,
                        )
                    )
                    continue
                current_token_count += line_token_count
                if current_token_count > tokens_per_split and line_buffer:
                    new_chunks.append(
                        MarkdownChunker._create_new_chunk_from_lines(
                            chunk.headers, line_buffer
                        )
                    )
                    line_buffer = []
                    current_token_count = initial_token_count + line_token_count
                line_buffer.append(line)
            if line_buffer:
                new_chunks.append(
                    MarkdownChunker._create_new_chunk_from_lines(
                        chunk.headers, line_buffer
                    )
                )

            return new_chunks

    @staticmethod
    def _create_new_chunk_from_lines(
        headers: list[MarkdownLine], lines: list[MarkdownLine]
    ) -> Chunk:
        """Utility function to create a chunk
        from a buffer of markdownLines.

        Args:
            headers (list[MarkdownLine]): the headers of the original chunk.
            lines (list[MarkdownLine]): The lines to put in the chunk.

        Returns:
            Chunk: the new chunk.
        """
        return Chunk(headers=headers, content=lines, start_line=lines[0].line_idx)

    @staticmethod
    def _split_line_into_chunks(
        line: MarkdownLine,
        line_token_count: int,
        hard_max_chunk_token_count: int,
        chunk_headers: list[MarkdownLine],
    ) -> list[Chunk]:
        """Splits a line that is too big to fit in a chunk
        into multiple lines, each producing a chunk.

        NOTE: to avoid the need of tokenizer.decode()
        we split be character.

        Args:
            line (MarkdownLine): the line to split.
            line_token_count (int): the token count of the line.
            hard_max_chunk_token_count (int): the maximum number of token a chunk can be.
            chunk_headers (list[MarkdownLine]): the headers of the chunk.
        """
        n_splits = line_token_count // hard_max_chunk_token_count + 1
        n_chars = len(line.text)
        n_chars_per_split = n_chars // n_splits + 1
        line_splits = [
            MarkdownLine(
                text=line.text[n_chars_per_split * i : n_chars_per_split * (i + 1)],
                line_idx=line.line_idx,  # WARNING : This creates multiple lines with same idx.
                isin_code_block=line.isin_code_block,
                page=line.page,
            )
            for i in range(n_splits)
        ]

        return [
            MarkdownChunker._create_new_chunk_from_lines(chunk_headers, [line_split])
            for line_split in line_splits
        ]



import re
from pathlib import Path

from ...core.components import MarkdownDoc
from ...core.custom_markdownify import (
    CustomMarkdownConverter,
)  # type: ignore : no stub file
from ..abstract_parser import AbstractParser


class HTMLParser(AbstractParser):

    def parse_string(self, string: str) -> MarkdownDoc:
        """Parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            string (str): the markdown formatted string

        Returns:
            MarkdownDoc: the parsed document. Can be fed to chunker.
        """
        formatted_string = HTMLParser.apply_markdownify(string)
        formatted_string = HTMLParser.cleanup_string(formatted_string)

        return MarkdownDoc.from_string(formatted_string)

    def parse_file(self, filepath: str) -> MarkdownDoc:
        """Reads and parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            filepath (FilePath): the path to a .html file

        Returns:
            MarkdownDoc: the parsed document. Can be fed to chunker.
        """
        html_string = HTMLParser.read_file(filepath)

        return self.parse_string(html_string)

    @staticmethod
    def read_file(filepath: str) -> str:
        """Reads a Markdown file

        Args:
            filepath (str): the path to the HTML file.

        Returns:
            str: the HTML string.
        """
        path = Path(filepath)
        if path.suffix != ".html":
            raise ValueError("Only .html files can be passed to HTMLParser.")
        with path.open("r", encoding="utf8") as file:
            html_string = file.read()

        return html_string

    @staticmethod
    def apply_markdownify(html_string: str) -> str:
        """Applies markdownify to the html text

        Args:
            html_string (str): an HTML-formatted string.

        Returns:
            str: the markdownified string.
        """
        md_string = html_string
        initial_len, new_len = len(md_string), 0
        while new_len < initial_len:
            initial_len = len(md_string)
            md_string = CustomMarkdownConverter(
                autolinks=False,
                heading_style="ATX",
                strip=["figure", "img"],
                bullets="-*+",
                escape_asterisks=False,
                escape_underscores=False,
                escape_misc=False,
            ).convert(  # type: ignore MarkdownConverter.convert()->str
                md_string
            )
            new_len = len(md_string)

        md_string = "\n".join(
            line + "\n" if line.startswith("#") else line
            for line in str(md_string).split("\n")
        )

        return md_string

    @staticmethod
    def cleanup_string(md_string: str) -> str:
        """Cleans up the html string.

        Args:
            md_string (str): the markdown string, output from
                apply_markdownify()

        Returns:
            str: the cleaned up string.
        """
        md_string = md_string.strip()
        md_string = re.sub(r"(?:\n\s*){3,}", "\n\n", md_string)

        # remove base64 images
        pattern = r"data:image\/[bmp,gif,ico,jpg,png,svg,webp,x\-icon,svg+xml]+;base64,[a-zA-Z0-9,+,\/]+={0,2}"
        md_string = re.sub(pattern, "", md_string)

        return md_string



from pathlib import Path
import re
from typing import Any

import yaml

from ...core.components import MarkdownDoc
from ..abstract_parser import AbstractParser


class MarkdownParser(AbstractParser):

    def parse_string(self, string: str) -> MarkdownDoc:
        """Parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            string (str): the markdown formatted string

        Returns:
            TypedString: the formatted markdown string
        """
        formatted_string = MarkdownParser.cleanup_string(string)
        formatted_string, metadata = MarkdownParser.parse_metadata(formatted_string)
        formatted_string = MarkdownParser.convert_setext_to_atx(formatted_string)
        md_doc = MarkdownDoc.from_string(formatted_string)
        md_doc.metadata = metadata

        return md_doc

    def parse_file(self, filepath: str) -> MarkdownDoc:
        """Reads and parses a markdown-formatted string.
        Ensures that the formatting is suited to be passed
        to the MarkdownChunker.

        Args:
            filepath (FilePath): the path to a .md file

        Returns:
            TypedString: the typed string
        """
        md_string = MarkdownParser.read_file(filepath)

        return self.parse_string(md_string)

    @staticmethod
    def read_file(filepath: str) -> str:
        """Reads a Markdown file

        Args:
            filepath (str): the path to the markdown file

        Returns:
            str: the markdown string
        """
        path = Path(filepath)
        if path.suffix != ".md":
            raise ValueError("Only .md files can be passed to MarkDownParser.")
        with path.open("r", encoding="utf8") as file:
            md_string = file.read()

        return md_string

    @staticmethod
    def convert_setext_to_atx(md_string: str) -> str:
        """Converts headers from setext style to atx style

        Args:
            md_string (str): the markdown string

        Return:
            str: the string with formatted headers
        """
        output_lines: list[str] = []
        within_code_block = False
        prev_line = ""

        for line in md_string.split("\n"):
            if line.startswith("```") and "```" not in line[3:]:
                within_code_block = not within_code_block
            if not within_code_block:
                if line.startswith("==="):
                    output_lines.pop()
                    output_lines.append(f"# {prev_line}")
                elif line.startswith("---"):
                    output_lines.pop()
                    output_lines.append(f"## {prev_line}")
                else:
                    output_lines.append(line)
            else:
                output_lines.append(line)
            prev_line = line

        return "\n".join(output_lines)

    @staticmethod
    def parse_metadata(md_string: str) -> tuple[str, dict[str, Any]]:
        """Parses the metadatas of a markdown string.
        Assumes the metadata are in YAML format, with '---' as first line.
        Example :
        ```md
        ---
        metakey : metavalue
        ---

        Content of document...
        ```
        Args:
            md_string (str): the string to get the metadata from

        Returns:
            str: the content of the docu, with the metadata section removed
            dict[str, Any]: the parsed metadata, as dict
        """
        if not md_string[:3] == "---":  # no metadata
            return md_string, {}
        # find idx of lines belonging to metadata
        lines = md_string.split("\n")
        try:
            end_idx = lines[1:].index("---") + 1
        except ValueError:
            return md_string, {}
        metadata = "\n".join(lines[1:end_idx]).strip()
        content = "\n".join(lines[end_idx + 1 :]).strip()
        try:
            metadata_as_json = yaml.safe_load(metadata) or {}
        except yaml.YAMLError:
            metadata_as_json = {}

        return content, metadata_as_json

    @staticmethod
    def cleanup_string(md_string: str) -> str:
        """Cleans up the html string.

        Args:
            md_string (str): the markdown string, output from
                apply_markdownify()

        Returns:
            str: the cleaned up string.
        """
        md_string = md_string.strip()
        md_string = re.sub(r"(?:\n\s*){3,}", "\n\n", md_string)

        # remove base64 images
        pattern = r"data:image\/[bmp,gif,ico,jpg,png,svg,webp,x\-icon,svg+xml]+;base64,[a-zA-Z0-9,+,\/]+={0,2}"
        md_string = re.sub(pattern, "", md_string)

        return md_string



from .abstract_parser import AbstractParser
from .docx.docx_parser import DocxParser
from .html.html_parser import HTMLParser
from .json.wikit_parser import WikitJsonParser
from .markdown.markdown_parser import MarkdownParser
from .notebook.jupyter_notebook_parser import JupyterNotebookParser
from .pdf.pdf_parser import PdfParser
from .sheets.csv_parser import CSVParser
from .sheets.excel_parser import ExcelParser



from itertools import groupby

import numpy as np
import numpy.typing as npt
import pymupdf  # type: ignore | no stub files

from .components import Link, TextSpan
from .utils import PdfParserState


class PdfLinkExtraction(PdfParserState):
    """Class intended to be used to extract the links from the document
    and bind them to the corresponding spans.
    Intended to be a component inherited by pdfParser => PdfParser(PdfLinkExtraction)
    """

    def _bind_links_to_spans(self, spans: list[TextSpan]) -> list[TextSpan]:
        """In a pdf, links are just an invisible clickable box
        layered on top of a span.
        This method gets the links of the pdf
        and binds them to their corresponding span.

        Args:
            spans (list[TextSpan]): the list of spans

        Returns:
            list[TextSpan]: the list of spans, with the "link" and "has_link" attributes updated
        """
        spans_per_page_map = {
            page_n: list(spans_on_page)
            for page_n, spans_on_page in groupby(spans, key=lambda span: span.page)
        }
        links_per_page_map: dict[int, list[Link]] = {
            page.number: [  # type: ignore | missing typing in pymupdf: Page.number : int
                Link(uri=link["uri"], bbox=link["from"])
                for link in page.links(kinds=(pymupdf.LINK_URI,))  # type: ignore | missing typing in pymupdf: Page.links() -> list[dict[str, str]]
            ]
            for page in self.document.pages(self.page_start, self.page_end)  # type: ignore | missing typing in pymupdf: Document.Pages() -> Generator(Page)
        }
        for page_n, links_on_page in links_per_page_map.items():
            if not links_on_page or not page_n in spans_per_page_map:
                continue  # No links to bind on that page, or no spans to bind links to
            links_bboxes = np.array([link.bbox for link in links_per_page_map[page_n]])
            spans_bboxes = np.array([span.bbox for span in spans_per_page_map[page_n]])
            intersection_areas = PdfLinkExtraction.calculate_intersection_areas(
                spans_bboxes, links_bboxes
            )  # 2D matrix of intersection areas. Shape : (n_spans, n_links)
            for i in range(intersection_areas.shape[1]):
                corresponding_span_idx = (
                    PdfLinkExtraction._get_span_corresponding_to_link(
                        intersection_areas[:, i]
                    )
                )
                if corresponding_span_idx is not None:
                    spans_per_page_map[page_n][corresponding_span_idx].link = (
                        links_per_page_map[page_n][i]
                    )

        return spans

    @staticmethod
    def calculate_intersection_areas(
        spans_bboxes: npt.NDArray[np.float32], links_bboxes: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        """Calculates the intersection areas of each span bbox
        respective to each link bbox.

        Args:
            spans_bboxes (npt.NDArray[np.float32]): A 2D array representing the bbox coordinates of all  the spans' bboxes.
                Expected to be of shape (n_spans_bboxes, 4).
            links_bboxes (npt.NDArray[np.float32]):  A 2D array representing the bbox coordinates of all the links' bboxes.
                Expected to be of shape (n_links_bboxes, 4).

        Returns:
            npt.NDArray[np.float32]: A 2D array of shape (n_links_bboxes, n_spans_bboxes) with each value
                representing the intersection area of the ith span's bbox with jth link's bbox.
        """
        # Extract the coordinates for the intersection calculation
        x11, y11, x12, y12 = (
            spans_bboxes[:, 0][:, None],
            spans_bboxes[:, 1][:, None],
            spans_bboxes[:, 2][:, None],
            spans_bboxes[:, 3][:, None],
        )
        x21, y21, x22, y22 = (
            links_bboxes[:, 0],
            links_bboxes[:, 1],
            links_bboxes[:, 2],
            links_bboxes[:, 3],
        )

        # Calculate the coordinates of the intersection rectangle
        x_left = np.maximum(x11, x21)
        y_top = np.maximum(y11, y21)
        x_right = np.minimum(x12, x22)
        y_bottom = np.minimum(y12, y22)

        # Calculate intersection areas
        inter_width = np.maximum(0, x_right - x_left)
        inter_height = np.maximum(0, y_bottom - y_top)
        intersection_areas = inter_width * inter_height

        return intersection_areas

    @staticmethod
    def _get_span_corresponding_to_link(
        intersect_areas: npt.NDArray[np.float16],
    ) -> int | None:
        """Given an array of intersect areas, find the idx of the spans
        that corresponds to the link. The array must be of shape (n_spans,).

        To find the span a link is bound to, we consider the area of the
        intersection of the link's bbox and the span's bbox.
        Two different situations:
        - 1) there is only one span with a high bbox intersection value with the link
            => this span is the one of the link
        - 2) there are multiple spans with a high bbox intersection.
            => for some reason, the span corresponding to the link is not the one
            with the biggest bbox intersection area, but the one with the second biggest.
        Args:
            intersect_areas (npt.NDArray[np.float16]): an array of intersections area of a link with
                all spans on the same page.

        WARNING : MAY return None if no span with bbox intersection with the link's bbox has been found

        Returns:
            int: the index of the corresponding span in the list
        """
        # Build a list of (span_idx, span/link_intersect_area) tuples, sorted by intersect area in descending order
        idx_area_tuples = list(
            zip(np.argsort(-intersect_areas), -np.sort(-intersect_areas))
        )

        idx_area_tuples = [item for item in idx_area_tuples if item[1] > 0]
        if not idx_area_tuples:
            return None  # no span with intersection with link

        best_idx, best_area = idx_area_tuples[0]
        if len(idx_area_tuples) == 1:
            return best_idx
        elif len(idx_area_tuples) > 1:
            if best_area / 2 > idx_area_tuples[1][1]:
                return best_idx
            else:
                idx_area_tuples = [
                    item for item in idx_area_tuples if item[1] < best_area
                ]
                if idx_area_tuples:
                    return idx_area_tuples[0][0]
                return None



from typing import Literal, Optional

from nbformat import NotebookNode
from pydantic import BaseModel, ConfigDict


class JupyterNotebookCell(BaseModel):
    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)
    cell_type: Literal["markdown", "code", "raw"]
    metadata: NotebookNode
    source: str
    execution_count: Optional[int | None] = None
    outputs: Optional[list[NotebookNode]] = None
    attachments: Optional[NotebookNode] = None


class JupyterNotebookContent(BaseModel):
    """Utility schemas intended to be used for the parsing of notebooks.
    Ensures the notebook formatting in ok. Details about the format a the notebook
    can be found in nbformat's documentation : https://nbformat.readthedocs.io/en/latest/format_description.html#
    """

    model_config = ConfigDict(populate_by_name=True, arbitrary_types_allowed=True)
    metadata: NotebookNode
    nbformat: int
    nbformat_minor: int
    cells: list[JupyterNotebookCell]



from __future__ import annotations

import json
import re
from copy import deepcopy
from typing import Any
from unicodedata import normalize

from pydantic import BaseModel, ConfigDict, Field, computed_field


class MarkdownDoc(BaseModel):
    """A parsed Markdown Formatted-String,
    resulting in a list of MarkdownLine.
    Feats :
    - ATX header formatting.
    - Remove base64 images
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    content: list[MarkdownLine]
    metadata: dict[str, Any] = {}

    def to_string(self) -> str:
        """Get the markdown string corresponding to the document's content"""
        return "\n".join(line.text for line in self.content)

    @staticmethod
    def from_string(md_string: str) -> MarkdownDoc:
        """Get the MardownDoc object from
        a markdown formatted string.

        Args:
            md_string (str): the markdown string

        Returns:
            MarkdownDoc: the markdown document
        """
        within_code_block = False
        md_lines: list[MarkdownLine] = []

        for i, line in enumerate(md_string.split("\n")):
            if line.startswith("```") and "```" not in line[3:]:
                within_code_block = not within_code_block
            md_lines.append(
                MarkdownLine(
                    text=line,
                    line_idx=i,
                    isin_code_block=(line == "```") or within_code_block,
                )
            )

        return MarkdownDoc(content=md_lines)


class MarkdownLine(BaseModel):
    text: str = Field(description="the text content of the line")
    line_idx: int = Field(description="the index of the line in the markdown string")
    isin_code_block: bool = Field(
        description="whether or not the line belongs to a code block"
    )
    page: int | None = Field(
        description="the page the line belongs to (if markdown comes from converted paginated document)"
    )

    def __init__(
        self,
        text: str,
        line_idx: int,
        isin_code_block: bool = False,
        page: int | None = None,
        **kwargs: dict[str, Any],
    ) -> None:
        super().__init__(
            text=normalize("NFKD", text.strip()),
            line_idx=line_idx,
            isin_code_block=isin_code_block,
            page=page,
            **kwargs,
        )

    @property
    def isin_table(self) -> bool:
        """whether or not the line belongs to a table"""
        return self.text.startswith("|")

    @property
    def is_header(self):
        return self.text.lstrip("- ").startswith("#") and not self.isin_code_block

    @property
    def is_bullet_point(self) -> bool:
        "whether or not the line is a bullet point"
        return self.text.startswith("- ")

    def get_header_level(self) -> int:
        """Gets the header level of this line (1-based)

        Raises:
            ValueError: if the line is not a header, raises an error

        Returns:
            int: the header level, h1 headers would return 1
        """
        if not self.is_header:
            raise ValueError("No header level as this line is not a header")
        else:
            text = self.text.lstrip("- ")
            return min(len(text) - len(text.lstrip("#")), 6)

    def __str__(self):
        return str(vars(self))


class Chunk(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    headers: list[MarkdownLine]
    content: list[MarkdownLine]
    start_line: int

    @computed_field
    @property
    def word_count(self) -> int:
        """Gets the amount of words in the chunk's content
        (headers not included)
        """
        text_content = "\n".join((line.text for line in self.content))
        return len(re.findall(r"\w+", Chunk._cleanup_text(text_content)))

    @computed_field
    @property
    def start_page(self) -> int | None:
        pages = [line.page for line in self.content if line.page is not None]
        if pages:
            return min(pages)
        else:
            return None

    @computed_field
    @property
    def end_page(self) -> int | None:
        pages = [line.page for line in self.content if line.page is not None]
        if pages:
            return max(pages)
        else:
            return None

    def __str__(self) -> str:
        return self.get_text()

    def get_text(self, remove_links: bool = False, prepend_headers: bool = True) -> str:
        """Gets the text of the chunk.

        Args:
            remove_links (bool, optional): If True, the markdown links will be removed (text of the link is kept).
                Defaults to False.

        Returns:
            str: the text
        """
        text = ""
        if prepend_headers:
            text += "\n\n".join((header.text for header in self.headers)) + "\n\n"
        text += "\n".join((line.text for line in self.content))
        if remove_links:
            text = Chunk.remove_links(text)

        return Chunk._cleanup_text(text)

    @staticmethod
    def remove_links(text: str) -> str:
        """Removes the markdown format of the links in the text.

        Args:
            text (str): the text to find the links in

        Returns:
            str: the formated text
        """
        pattern = r"\[?!?\[(.*?)\](\(.*\)\])?\((.*\..+?)(\s.*)?\)"
        matches = re.finditer(pattern, text)
        for match in matches:
            text = text.replace(match[0], match[1])

        return text

    @staticmethod
    def _cleanup_text(text: str) -> str:
        """Cleans up the text"""
        text = re.sub(r"\n{3,}", "\n\n", text)
        text = text.strip()

        return text


class TocTree:
    id: int
    title: MarkdownLine
    content: list[MarkdownLine]
    parent: TocTree | None
    children: list[TocTree]

    def __init__(
        self,
        title: MarkdownLine,
        *,
        content: list[MarkdownLine] | None = None,
        children: list[TocTree] | None = None,
        id: int = -1,
        parent: TocTree | None = None,
    ) -> None:
        self.title = title
        self.content = [] if content is None else content
        self.id = id
        self.parent = parent
        self.children = [] if children is None else children

    def add_child(self, child: TocTree) -> None:
        """Adds a child to the list of TocTree.
        Self is added as parent

        Args:
            child (TocTree): a TocTree node to add to the children.
        """
        child.parent = self
        self.children.append(child)

    def get_title_by_id(self, id: int) -> TocTree | None:
        """Gets a toc tree using its id.

        Args:
            id (int): the id of the title we are looking for.

        Returns:
            TocTree: the element of title we were looking for
        """
        if self.id == id:
            return self
        for child in self.children:
            target = child.get_title_by_id(id)
            if target:
                return target

    def to_json(self, output_path: str = "./tree.json") -> None:
        """Outputs the tree as json.

        Returns:
            dict[str, Any]: the json
        """
        dumpable_tree = deepcopy(self)
        dumpable_tree.remove_circular_refs()
        with open(output_path, "w", encoding="utf-8") as file:
            json.dump(
                dumpable_tree,
                file,
                default=lambda o: o.__dict__,
                sort_keys=True,
                ensure_ascii=False,
                indent=4,
            )

    def remove_circular_refs(self):
        """Recursively removes the circular ref in dict
        (used to save as json).

        Args:
            tree (TocTree): a toc tree to remove circular refs
        """
        self.parent = None
        for child in self.children:
            child.remove_circular_refs()

    def get_text(self, content_only: bool = False) -> str:
        """Builds the text content of a toc tree

        Args:
            content_only (bool): if true, the text won't include headers, only content.

        Returns:
            str : the text content
        """
        text = "" if content_only else self.title.text + "\n\n"
        text += "\n".join((line.text for line in self.content))

        return text.strip()



from .components import Link, TextBlock, TextLine, TextSpan
from .components_tables import Cell, PdfTable, TableFinder
from .export import PdfExport
from .extract_links import PdfLinkExtraction
from .extract_tables import PdfTableExtraction
from .extract_toc import PdfTocExtraction
from .plot import PdfPlotter
from .utils import DocSpecsExtraction, PdfParserState



from collections import Counter
from typing import Literal

import pymupdf  # type: ignore : no stubs

from ....exceptions.exceptions import PdfParserException
from .components import TextBlock, TextLine, TextSpan
from .components_tables import PdfTable, TableFinder


class PdfParserState:
    """Contains the state of the class.
    Intended to be inherited by all pdfParser components
    """

    _document: pymupdf.Document | None = None
    filepath: str | None = None
    page_start: int = 0
    page_end: int | None = None
    spans: list[TextSpan] = []
    lines: list[TextLine] = []
    blocks: list[TextBlock] = []
    tables: list[PdfTable] = []
    table_finder: TableFinder = TableFinder()
    main_title: str = ""
    main_body_fontsizes: list[float] = []
    document_fontsizes: list[float] = []
    main_body_is_bold: bool = False
    document_orientation: Literal["portrait", "landscape"]

    @property
    def document(self) -> pymupdf.Document:
        if self._document is None:
            raise PdfParserException(
                "No document parsed. You might call parser.parse_file('my_file.pdf')."
            )
        return self._document

    @document.setter
    def document(self, doc: pymupdf.Document) -> None:
        self._document = doc


class DocSpecsExtraction(PdfParserState):

    def _set_document_specifications(self) -> None:
        """Set the specifications of various specifications of the document.
        Stores the attributes in :
        - self.document_orientation -> the orientation of the document (portrait or landscape)
        - self.main_body_fontsizes -> the fontsizes used for the body content of the document
        - self.document_fonsizes -> a sorted list of fontsizes in the document (bigger than body)
        - self.main_body_is_bold -> whether or not the main body is written in bold

        Meant to be called after the creation of the TextBlock objects.
        """
        self._set_document_font_specs()
        self._set_document_orientation()

    def _set_document_orientation(self) -> None:
        """Determines the orientation of the document
        using the first page of the document.

        - self.document_orientation -> the orientation of the document (portrait or landscape)
        """
        page_rect: pymupdf.Rect = self.document[0].rect
        self.document_orientation = "portrait" if page_rect.height > page_rect.width else "landscape"  # type: ignore missing typing in pymupdf | Rect.height : float, Rect.width : float

    def _set_document_font_specs(self):
        """Set the specifications of various specifications
        regarding the fonts in the document.
        Stores the attributes in :
        - self.main_body_fontsizes -> the fontsizes used for the body content of the document
        - self.document_fonsizes -> a sorted list of fontsizes in the document (bigger than body)
        - self.main_body_is_bold -> whether or not the main body is written in bold
        """
        fontsize_counts = Counter(
            line.fontsize
            for line in self.lines
            if not line.is_empty and line.orientation == (1.0, 0.0)
        )
        if not fontsize_counts:
            return
        self.document_fontsizes = list(fontsize_counts.keys())
        self.main_body_fontsizes = [
            fontsize
            for fontsize, occurence in fontsize_counts.items()
            if occurence > len(self.lines) * 0.1
        ]
        if not self.main_body_fontsizes:
            return
        # determine whether or not the body of document is written in bold (if 30% of the lines are bold, we consider the main body is bold)
        bold_main_body_lines = [
            line.is_bold
            for line in self.lines
            if line.fontsize in self.main_body_fontsizes and not line.is_empty
        ]
        self.main_body_is_bold = (
            sum(bold_main_body_lines) / len(bold_main_body_lines) > 0.3
        )



import json
import os
from abc import ABC, abstractmethod
from typing import Any

from ..core.components import Chunk


class AbstractChunker(ABC):

    @abstractmethod
    def chunk(self, content: Any) -> list[Chunk]:
        """Considering the output of a specified parser,
        must contain all the logic for chunking a string.

        Args:
            content (Any): the content to chunk. Might correspond to the
                output from the provided parser.

        Returns:
            list[Chunk]: A list of chunks
        """
        pass

    def __call__(self, string: Any) -> list[Chunk]:
        """Shortcut to call chunk() method."""
        return self.chunk(string)

    def save_chunks(
        self, chunks: list[Chunk], output_filename: str, remove_links: bool = False
    ) -> None:
        """Saves the chunks in .md files at the designated location.

        Args:
            chunks (Chunks): the chunks.
            output_filename (str): the JSON file where to save the files. Must be json.
            remove_links (bool): Whether or not links should be remove from the chunk's text content.
        """
        directory = os.path.dirname(os.path.abspath(output_filename))
        if not os.path.exists(directory):
            os.makedirs(directory)
        content = [
            {
                "text": chunk.get_text(remove_links=remove_links),
                "start_page": chunk.start_page,
                "end_page": chunk.end_page,
                "start_line": chunk.start_line,
            }
            for chunk in chunks
        ]
        with open(output_filename, "w", encoding="utf8") as f:
            json.dump(content, f, ensure_ascii=False, indent=4)
